from ursina import *; matcap_shader = Shader(language=Shader.GLSL, fragment='''
#version 130

/*
* Copyright (C) 2003 Maxim Stepin ( maxst@hiend3d.com )
*
* Copyright (C) 2010 Cameron Zemek ( grom@zeminvaders.net )
*
* Copyright (C) 2014 Jules Blok ( jules@aerix.nl )
*
* This program is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#define SCALE 4

#if defined(VERTEX)

attribute vec4 VertexCoord;
attribute vec4 TexCoord;

uniform mat4 MVPMatrix;
uniform int FrameDirection;
uniform int FrameCount;
uniform vec2 OutputSize;
uniform vec2 TextureSize;
uniform vec2 InputSize;

varying vec4 vTexCoord[4];

void main()
{
	gl_Position = VertexCoord * transpose(MVPMatrix);

	vec2 ps = 1.0/TextureSize;
	float dx = ps.x;
	float dy = ps.y;

	//   +----+----+----+
	//   |    |    |    |
	//   | w1 | w2 | w3 |
	//   +----+----+----+
	//   |    |    |    |
	//   | w4 | w5 | w6 |
	//   +----+----+----+
	//   |    |    |    |
	//   | w7 | w8 | w9 |
	//   +----+----+----+

	vTexCoord[0].zw = ps;
	vTexCoord[0].xy = TexCoord.xy;
	vTexCoord[1] = TexCoord.xxxy + vec4(-dx, 0, dx, -dy); //  w1 | w2 | w3
	vTexCoord[2] = TexCoord.xxxy + vec4(-dx, 0, dx,   0); //  w4 | w5 | w6
	vTexCoord[3] = TexCoord.xxxy + vec4(-dx, 0, dx,  dy); //  w7 | w8 | w9
}


uniform sampler2D Texture;
uniform sampler2D LUT;
uniform int FrameDirection;
uniform int FrameCount;
uniform vec2 OutputSize;
uniform vec2 TextureSize;
uniform vec2 InputSize;

varying vec4 vTexCoord[4];

const mat3 yuv_matrix = mat3(0.299, 0.587, 0.114, -0.169, -0.331, 0.5, 0.5, -0.419, -0.081);
const vec3 yuv_threshold = vec3(48.0/255.0, 7.0/255.0, 6.0/255.0);
const vec3 yuv_offset = vec3(0, 0.5, 0.5);

bool diff(vec3 yuv1, vec3 yuv2)
{
	bvec3 res = greaterThan(abs((yuv1 + yuv_offset) - (yuv2 + yuv_offset)), yuv_threshold);
	return res.x || res.y || res.z;
}

void main()
{
	vec2 fp = fract(vTexCoord[0].xy*TextureSize);
	vec2 quad = sign(-0.5 + fp);
	mat3 yuv = transpose(yuv_matrix);

	float dx = vTexCoord[0].z;
	float dy = vTexCoord[0].w;
	vec3 p1  = texture2D(Texture, vTexCoord[0].xy).rgb;
	vec3 p2  = texture2D(Texture, vTexCoord[0].xy + vec2(dx, dy) * quad).rgb;
	vec3 p3  = texture2D(Texture, vTexCoord[0].xy + vec2(dx, 0) * quad).rgb;
	vec3 p4  = texture2D(Texture, vTexCoord[0].xy + vec2(0, dy) * quad).rgb;
	mat4x3 pixels = mat4x3(p1, p2, p3, p4);

	vec3 w1  = yuv * texture2D(Texture, vTexCoord[1].xw).rgb;
	vec3 w2  = yuv * texture2D(Texture, vTexCoord[1].yw).rgb;
	vec3 w3  = yuv * texture2D(Texture, vTexCoord[1].zw).rgb;

	vec3 w4  = yuv * texture2D(Texture, vTexCoord[2].xw).rgb;
	vec3 w5  = yuv * p1;
	vec3 w6  = yuv * texture2D(Texture, vTexCoord[2].zw).rgb;

	vec3 w7  = yuv * texture2D(Texture, vTexCoord[3].xw).rgb;
	vec3 w8  = yuv * texture2D(Texture, vTexCoord[3].yw).rgb;
	vec3 w9  = yuv * texture2D(Texture, vTexCoord[3].zw).rgb;

	bvec3 pattern[3];
	pattern[0] =  bvec3(diff(w5, w1), diff(w5, w2), diff(w5, w3));
	pattern[1] =  bvec3(diff(w5, w4), false       , diff(w5, w6));
	pattern[2] =  bvec3(diff(w5, w7), diff(w5, w8), diff(w5, w9));
	bvec4 cross = bvec4(diff(w4, w2), diff(w2, w6), diff(w8, w4), diff(w6, w8));

	vec2 index;
	index.x = dot(vec3(pattern[0]), vec3(1, 2, 4)) +
			  dot(vec3(pattern[1]), vec3(8, 0, 16)) +
			  dot(vec3(pattern[2]), vec3(32, 64, 128));
	index.y = dot(vec4(cross), vec4(1, 2, 4, 8)) * (SCALE * SCALE) +
	          dot(floor(fp * SCALE), vec2(1, SCALE));

	vec2 step = 1.0 / vec2(256.0, 16.0 * (SCALE * SCALE));
	vec2 offset = step / 2.0;
	vec4 weights = texture2D(LUT, index * step + offset);
	float sum = dot(weights, vec4(1));
	vec3 res = pixels * (weights / sum);

	gl_FragColor.rgb = res;
}

#endif
''',
)



if __name__ == '__main__':
    from ursina import *
    from ursina.prefabs.primitives import *
    app = Ursina(borderless=False)
    window.color=color.black

    shader = matcap_shader
    Texture.default_filtering = 'bilinear'
    a = Entity(model='quad', shader=shader, texture='brick')
    EditorCamera()

    app.run()



import sys
sys.path.append("..")
from ursina import *

class LoadModelButton():

    def __init__(self):
        self.path = None

    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            if self.path:
                print(self.path)
                self.load_model()

    @undoable
    def load_model(self):
        entity = Entity()
        entity.name = os.path.basename(self.path).split('.')[0]
        entity.model = os.path.basename(self.path).split('.')[0]

        entity.collision = True
        entity.editor_collider = 'box'
        button_script = entity.add_script('editor_draggable')
        scene.editor.hierarchy_panel.populate()

        # undo
        yield 'Load Model ' + entity.name
        destroy(entity)
        scene.editor.hierarchy_panel.populate()



from ursina import *
from direct.stdpy import thread


class LoadingWheel(Entity):
    def __init__(self, **kwargs):
        super().__init__()
        self.parent = camera.ui
        self.point = Entity(parent=self, model=Circle(24, mode='point', thickness=.03), color=color.light_gray, y=.75, scale=2, texture='circle')
        self.point2 = Entity(parent=self, model=Circle(12, mode='point', thickness=.03), color=color.light_gray, y=.75, scale=1, texture='circle')

        self.scale = .025
        self.text_entity = Text(world_parent=self, text='loading...', origin=(0,1.5), color=color.light_gray)
        self.y = -.25

        self.bg = Entity(parent=self, model='quad', scale_x=camera.aspect_ratio, color=color.black, z=1)
        self.bg.scale *= 400

        for key, value in kwargs.items():
            setattr(self, key ,value)


    def update(self):
        self.point.rotation_y += 5
        self.point2.rotation_y += 3



if __name__ == '__main__':
    app = Ursina()
    window.color = color.white
    info_text = Text('''Press space to start loading textures''', origin=(0,0), color=color.black)
    loading_screen = LoadingWheel(enabled=False)
    from ursina.prefabs.health_bar import HealthBar

    def load_textures():
        textures_to_load = ['brick', 'shore', 'grass', 'heightmap'] * 50
        bar = HealthBar(max_value=len(textures_to_load), value=0, position=(-.5,-.35,-2), scale_x=1, animation_duration=0, world_parent=loading_screen, bar_color=color.gray)
        for i, t in enumerate(textures_to_load):
            load_texture(t)
            print(i)
            bar.value = i+1
        # destroy(bar, delay=.01)
        print('loaded textures')
        loading_screen.enabled = False

    def input(key):
        if key == 'space':
            loading_screen.enabled = True
            info_text.enabled = False
            t = time.time()

            try:
                thread.start_new_thread(function=load_textures, args='')
            except Exception as e:
                print('error starting thread', e)

            print('---', time.time()-t)
    # load_textures()
    # invoke(load_textures, delay=0)
    app.run()



from ursina import *

# def play_synth_adsr(
#     wave = 'sine',
#     max_volume = .75,
#     # start_volume = 0,
#
#     attack = .05,
#     decay = .01,             # time to fade down to sustain volume
#     sustain_volume = .5,     # volume of middle part relative to max volume
#     sustain_length = .2,     # duration of middle part
#     release = .2,            # fade out duration
#     # release_volume = None,   # default is the same as sustain
#     # end_volume = 0,
#
#     start_pitch = 0,
#     end_pitch = 2,
#     pitch_curve = curve.linear,
#     # volume_multiplier = 1
#
#         ):
#
#     total_duration = attack + decay + sustain_length + release
#     # if release_volume is None:
#     #     release_volume = sustain_volume
#
#     a = Audio(wave, loop=True, pitch=pow(1 / 1.05946309436, -pitch), volume=start_volume)
#     a.animate('volume', max_volume, duration=attack, curve=curve.linear)
#     a.animate('volume', sustain_volume*max_volume, duration=decay, delay=attack, curve=curve.linear)
#     a.animate('volume', sustain_volume*max_volume, duration=sustain_length, delay=attack+decay, curve=curve.linear)
#     a.animate('volume', 0, duration=release, delay=attack+decay+sustain_length, curve=curve.linear)
#
#     a.animate('pitch', pow(1 / 1.05946309436, -end_pitch), duration=total_duration, curve=pitch_curve)
#     a.animations.append(invoke(a.stop, delay=total_duration))
#     return a


def ursfx(volume_curve, volume=.75, wave='sine', pitch=0, pitch_change=0, speed=1, pitch_curve=curve.linear, ignore_paused=False):  # play a retro style sound effect
    a = Audio(wave, loop=True, pitch=pow(1 / 1.05946309436, -pitch), volume=volume_curve[0][1] * volume, ignore_paused=ignore_paused)

    for i in range(len(volume_curve)-1):
        invoke(a.animate, 'volume', volume_curve[i+1][1] * volume, duration=(volume_curve[i+1][0] - volume_curve[i][0]) / speed, curve=curve.linear, unscaled=True, delay=volume_curve[i][0]/speed)

    a.animate('pitch', pow(1 / 1.05946309436, -pitch-pitch_change), duration=volume_curve[i-1][0] / speed, curve=pitch_curve, unscaled=True)
    a.animations.append(invoke(a.stop, delay=volume_curve[4][0] / speed, unscaled=True))

    invoke(a.stop, delay=volume_curve[4][0] / speed, unscaled=True, ignore_paused=ignore_paused)
    return a


class UrsfxGUI(Entity):
    def __init__(self, ignore_paused=True, play_after_change=True, **kwargs):
        super().__init__(**(dict(parent=camera.ui, z=-998, ignore_paused=ignore_paused) | kwargs))

        default_positions = [(0,0), (.1,.9), (.15,.75), (.6,.75), (1,0)]
        self.play_after_change = play_after_change
        self.wave_panel = Entity(parent=self, scale=.35, x=-0)
        self.waveform = Entity(parent=self.wave_panel, scale_y=.75)
        self.waveform_bg = Entity(parent=self.waveform, model='quad', origin=(-.5,-.5), z=.01, color=color.black66)
        # self.bg_2 = Entity(parent=self.waveform, model='quad', origin=(-.5,-.5), z=.01, color=color._16)
        self.volume_slider = Slider(parent=self.wave_panel, x=-.05, vertical=True, scale=1.95, min=.05, max=1, default=.75, step=.01, on_value_changed=self.play)

        self.wave_selector = ButtonGroup(('sine', 'triangle', 'square', 'noise'), parent=self.wave_panel, scale=.11, y=-.075)

        self.pitch_slider = Slider(parent=self.wave_panel, y=-.25, scale=1.95, min=-36, max=36, default=0, step=1, on_value_changed=self.play_updated, text='pitch')
        self.pitch_change_slider =  Slider(parent=self.wave_panel, y=-.325, scale=1.95, min=-12, max=12, default=0, step=1, on_value_changed=self.play_updated, text='pitch change')
        self.speed_slider =  Slider(parent=self.wave_panel, scale=1.95, min=.5, max=4, default=1, step=.1, on_value_changed=self.play_updated, y=-.4, text='speed')

        def coin_sound():
            self.paste_code(f"ursfx([(0,1), ({random.uniform(.05,.15)},.5), (.25,.5), ({random.uniform(.25,.5)},.5), (1,0)], pitch={random.randint(-3, 24)}, speed={random.uniform(2.5,3)})")
            self.play()

        self.coin_button = Button(text='coin', parent=self.wave_panel, scale=(.25, .125), origin=(-.5,.5), y=-.45, on_click=coin_sound)

        self.knobs = [Draggable(parent=self.waveform, scale=.05, model='circle', color=color.light_gray, highlight_color=color.azure, position=default_positions[i], i=i, min_y=0, max_y=1) for i in range(5)]
        self.knobs[0].lock = (True,False,True)

        self.line = Entity(parent=self.waveform, model=Mesh(vertices=[Vec3(0,0,0), Vec3(1,0,0)], mode='line', thickness=3), z=.01, color=color.light_gray)
        self.bg = Entity(parent=self.wave_panel, model='wireframe_quad', origin=(-.5,-.5), z=.02, color=color.black, scale_x=1)

        self.play_button = Button(text='>', parent=self.wave_panel, model='circle', scale=(.125, .125), color=color.azure, origin=(-.5,-.5), position=(-.075,1.025), on_click=self.play)
        self.copy_button = Button(text='copy', parent=self.wave_panel, scale=(.25, .125), color=color.dark_gray, origin=(-.5,-.5), position=(-.075+.125+.025, 1.025), on_click=self.copy_code)
        self.copy_button.text_entity.scale *= .5

        self.paste_button = Button(text='paste', parent=self.wave_panel, scale=(.25, .125), color=color.dark_gray, origin=(-.5,-.5), position=(self.copy_button.x + self.copy_button.scale_x + .025, 1.025), on_click=self.paste_code)
        self.paste_button.text_entity.scale *= .5

        self.code_text = Text('', parent=self.wave_panel, scale=2, position=(self.paste_button.x+self.paste_button.scale_x+.025,1.11))
        self.wave_selector.on_value_changed = self.play_updated
        self.background_panel = Entity(model=Quad(radius=.025), parent=self.wave_panel, color=color.black66, z=1, origin=(-.5,-.5), scale=(1.125,1.75+.025), position=(-.1,-.6))

        for i, knob in enumerate(self.knobs):
            def drag(this_knob=knob):
                if this_knob.i == 0:
                    this_knob.min_x = 0
                else:
                    print(self.knobs[this_knob.i - 1].x)
                    this_knob.min_x = self.knobs[this_knob.i - 1].x + .01

                if this_knob.i == 4:
                    this_knob.max_x = 3
                else:
                    this_knob.max_x = self.knobs[this_knob.i + 1].x - .01

            knob.drag = drag

            def drop(this_knob=knob):
                self.play()
                self.bg.scale_x = self.knobs[4].x

                for e in self.knobs:
                    e.min_x = -0
                    e.max_x = 3

            knob.drop = drop

        self.draw()

        for e in scene.entities:
            if e.has_ancestor(self):
                e.ignore_paused = ignore_paused


    def update(self):
        for e in self.knobs:
            e.world_scale = .25
        if sum(int(e.dragging) for e in self.knobs) == 0:
            return

        self.draw()


    def input(self, key):
        if held_keys['control'] and key == 'v' and self.wave_panel.enabled:
            self.paste_code()


    def copy_code(self):
        print(self.recipe)
        import pyperclip
        pyperclip.copy(self.recipe)


    def paste_code(self, code=""):
        import pyperclip
        if not code:
            code = pyperclip.paste()

        print('paste code:', code)
        try:
            curve = code.split('ursfx(')[1].split(']')[0] + ']'
            curve = eval(curve)
            for i, e in enumerate(curve):
                self.knobs[i].position=(e)
            print('--------curve', curve)

            volume = .75
            if 'volume=' in code:
                volume = code.split('volume=')[1].split(',')[0]
                volume = eval(volume)
                self.volume_slider.value = volume
                print('--------volume', volume)

            wave = 'sine'
            if 'wave=' in code:
                wave = code.split('wave=')[1].split(',')[0]
                wave = eval(wave)
                self.wave_selector.value = wave
                print('--------wave', wave)

            pitch = 1
            if 'pitch=' in code:
                pitch = code.split('pitch=')[1]
                if ',' in pitch:
                    pitch = pitch.split(',')[0]
                else:
                    pitch = pitch[:-1]
                pitch = eval(pitch)
            self.pitch_slider.value = pitch
            print('--------pitch', pitch)


            pitch_change = 0
            if 'pitch_change=' in code:
                pitch_change = code.split('pitch_change=')[1]
                if ',' in pitch_change:
                    pitch_change = pitch_change.split(',')[0]
                else:
                    pitch_change = pitch_change[:-1]
                pitch_change = eval(pitch_change)
            self.pitch_change_slider.value = pitch_change
            print('--------pitch_change')

            speed = 1
            if 'speed=' in code:
                speed = code.split('speed=')[1]
                if ',' in speed:
                    speed = speed.split(',')[0]
                else:
                    speed = speed[:-1]
                speed = eval(speed)
            self.speed_slider.value = speed
            print('--------speed')


            print(curve, volume, wave, pitch, pitch_change, speed)
            self.draw()

        except:
            print('invalid ursfx code:', code)


    @property
    def recipe(self):
        pitch_code = ''
        if self.pitch_slider.value:
            pitch_code = f', pitch={self.pitch_slider.value}'

        pitch_change_code = ''
        if self.pitch_change_slider.value:
            pitch_change_code = f', pitch_change={self.pitch_change_slider.value}'

        speed_code = ''
        if speed_code != 1:
            speed_code = f', speed={round(self.speed_slider.value, 1)}'

        return f"ursfx({str(self.volume_curve)}, volume={round(self.volume_slider.value,3)}, wave='{self.wave_selector.value}'{pitch_code}{pitch_change_code}{speed_code})"






    def draw(self):
        # self.bg.scale_x = self.knobs[4].x
        self.waveform_bg.scale_x = self.knobs[4].x
        self.bg.scale_x = self.knobs[4].x

        for e in self.knobs:
            e.position = (round(e.x,2), round(e.y,2))


        self.line.model.vertices = [e.get_position(relative_to=self.waveform) for e in self.knobs]
        self.line.model.generate()

        self.volume_curve = [e.get_position(relative_to=self.waveform) for e in self.knobs]
        self.volume_curve = [(round(e[0],2), round(e[1],2)) for e in self.volume_curve]

        self.waveform.scale_y = self.volume_slider.value
        # self.code_text.text = f"ursfx({str(self.volume_curve)}, {round(self.volume_slider.value,3)}, wave='{self.wave_selector.value}', pitch={self.pitch_slider.value}, end_pitch={self.pitch_slider.value+self.pitch_change_slider.value})"


    def play(self):
        self.draw()
        # ursfx([(0.0, 0.0), (0.02, 0.89), (0.08, 0.63), (0.15, 0.27), (0.19, 0.0)], volume=0.75, wave='noise', pitch=-12, pitch_change=-9)
        # select = ursfx([(0.0, 0.0), (0.03, 0.9), (0.04, 0.75), (0.08, 0.25), (0.2, 0.0)], volume=0.75, wave='triangle', pitch=2, pitch_change=12)
        # select_2 = ursfx([(0.0, 0.0), (0.03, 0.9), (0.04, 0.75), (0.07, 0.18), (0.25, 0.0)], volume=0.86, wave='triangle', pitch=1, pitch_change=7)
        # hurt = ursfx([(0.0, 0.0), (0.03, 0.9), (0.04, 0.75), (0.06, 0.25), (0.22, 0.0)], volume=0.9, wave='noise', pitch=-12, pitch_change=-12)
        # heal = ursfx([(0.0, 0.0), (0.04, 0.9), (0.08, 0.04), (0.13, 0.66), (0.45, 0.0)], volume=1.0, wave='sine', pitch=-2, pitch_change=4)
        # coind = ursfx([(0.0, 1.0), (0.07, 0.5), (0.25, 0.5), (0.39, 0.5), (1.0, 0.0)], volume=0.75, wave='sine', pitch=24, speed=2.7)
        ursfx(
            [e.get_position(relative_to=self.waveform) for e in self.knobs],
            volume=self.volume_slider.value,
            wave=self.wave_selector.value,
            pitch=self.pitch_slider.value,
            pitch_change=self.pitch_slider.value + self.pitch_change_slider.value,
            speed=self.speed_slider.value,
            ignore_paused=self.ignore_paused,
            )

        self.code_text.text = self.recipe

    def play_updated(self):
        if self.play_after_change:
            self.play()


gui = None
def open_gui():
    global gui
    if not gui:
        gui = UrsfxGUI(enabled=False)

        def toggle_gui_input(key):
            if key == 'f3':
                gui.enabled = not gui.enabled
        Entity(input=toggle_gui_input)
    gui.enabled = True


if __name__ == '__main__':
    app = Ursina()
    sfx_editor = UrsfxGUI()
    # application.paused = True
    Sprite('shore', z=10, ppu=64, color=color.gray)
    app.run()



from ursina import *
from ursina.shaders import triplanar_shader


app = Ursina()

for z in range(0,1000, 100):
    e = Entity(model='rock_shape', shader=triplanar_shader, scale=32, texture='magic_tree_ground')
    e.set_shader_input('top_texture', load_texture('grass')._texture)

    e.z = z
    e.x = random.uniform(-100,100)
plane = Entity(model='plane', scale=9999, texture='grass')
scene.fog_density = .001
EditorCamera()
Sky(texture='sky_sunset')
app.run()



from ursina import *


class Capsule(Mesh):
    def __init__(self, height=2, radius=.5, **kwargs):
        # make a capsule by stretching a sphere
        sphere_mesh = load_model('sphere', application.internal_models_compressed_folder, use_deepcopy=True)
        vertices = [Vec3(*v) + (Vec3(0,height-1,0) * (int(v[1] > 0)-.5)) for v in sphere_mesh.vertices]

        super().__init__(vertices=vertices, triangles=sphere_mesh.triangles, uvs=sphere_mesh.uvs, normals=sphere_mesh.normals, colors=sphere_mesh.colors, **kwargs)


if __name__ == '__main__':
    app = Ursina()
    e = Entity(model=Capsule(), texture='brick')
    origin = Entity(model='quad', color=color.orange, scale=(.05, .05), always_on_top=True)
    ed = EditorCamera(rotation_speed = 200, panning_speed=200)
    app.run()



from ursina import *
from panda3d.bullet import BulletWorld, BulletPlaneShape, BulletBoxShape, BulletRigidBodyNode, BulletDebugNode
from panda3d.core import *
from ursina.shaders import basic_lighting_shader

app = Ursina()

ground = Entity(model='plane', scale=10, texture='grass')
box = Entity(model='cube', color=color.yellow, shader=basic_lighting_shader)

world = BulletWorld()
world.setGravity(Vec3(0, -9.81, 0))


# def Entity():
#     node = BulletRigidBodyNode('Ground')
#     node.addShape(shape)
#     world.attachRigidBody(node)
#     return render.attachNewNode(node)
#

shape = BulletPlaneShape(Vec3(0, 1, 0), 0)
node = BulletRigidBodyNode('Ground')
node.addShape(shape)
np = render.attachNewNode(node)
ground.parent = np
world.attachRigidBody(node)

class RB(NodePath):
    def __init__(self, **kwargs):
        super().__init__('test')
        box_node = BulletRigidBodyNode('Box')
        box_node.setMass(1)

RB()

shape = BulletBoxShape(Vec3(0.5, 0.5, 0.5))
box_node = BulletRigidBodyNode('Box')
box_node.setMass(1.0)
box_node.addShape(shape)
world.attachRigidBody(box_node)

thing = NodePath(box_node)
thing.reparentTo(render)
thing.set_y(2)
# box.model.reparentTo(thing)
# box_node.reparentTo(thing)
destroy(box)
# thing.clear()
# thing.attachNewNode(box_node)
print('---------', thing.get_nodes())
# box. y = 2
# box_np = NodePath()
# box_np.attachNewNode(box_node)
# box_np = render.attachNewNode(box_node)
# box_np.setY(4)
# box.parent = box_np



Sky()
ed = EditorCamera()
ed.rotation = (20, 10, 0)
ed.y = 1
window.size *= .5
Text('Hold space to simulate physics', position=window.top, origin=(0,1), color=color.black)
debugNode = BulletDebugNode('Debug')
debugNode.showWireframe(True)
debugNode.showConstraints(True)
debugNode.showBoundingBoxes(False)
debugNode.showNormals(False)
debugNP = render.attachNewNode(debugNode)
debugNP.show()
world.setDebugNode(debugNP.node())

def update():
    if held_keys['space']:
        world.doPhysics(time.dt)
        print(thing.node())


app.run()



from ursina import *


e = Entity(parent=camera.ui, model='quad', texture='desert_terrain_heightmap')
t = e.texture


for y in range(t.height):
    for x in range(t.width):
        col = tex.get_pixel(x+1, y)
        



from pathlib import Path
from textwrap import dedent

import sswg
from documentation_parser import VarInfo, analyze_module, get_all_modules_in_package, get_package_version

import ursina
from ursina import application, camel_to_snake, flatten_list, print_warning, snake_to_camel


def generate_documentation():
    groups = {
        'Basics' : [
            'Ursina',
            'Entity',
            'Button',
            'Sprite',
            'Text',
            'Audio',
            ],
        'Core Modules' : [
            'camera',
            'mouse',
            'window',
            'application',
            'scene',
            ],
        'Graphics' : [
            'color',
            'Mesh',
            'Shader',
            'Texture',
            'Light',
            'DirectionalLight',
            'PointLight',
            'AmbientLight',
            'SpotLight',
            ],
        'Procedural Models': [
            'Quad',
            'Circle',
            'Plane',
            'Grid',
            'Cone',
            'Cylinder',
            'Pipe',
            'Terrain',
            ],
        'modules': [
            'input_handler',
            'mesh_importer',
            'texture_importer',
            'string_utilities',
            ],
        'Animation': [
            'Animation',
            'FrameAnimation3d',
            'SpriteSheetAnimation',
            'Animator',
            'TrailRenderer',
            'curve',
            ],
        'Math': [
            'ursinamath',
            'Vec2',
            'Vec3',
            'Vec4',
            'CubicBezier',
            'array_tools',
            'Array2D',
            'Array3D',
        ],
        'Gameplay' : [
            'ursinastuff',
            'Sequence',
            'Func',
            'Keys',
        ],
        'Collision' : [
            'raycast',
            'terraincast',
            'boxcast',
            'HitInfo',
            'Collider',
            'BoxCollider',
            'SphereCollider',
            'MeshCollider',
        ],
        'Prefabs': [
            'Sky',
            'EditorCamera',
            'Tilemap',
            'FirstPersonController',
            'PlatformerController2d',
            'Conversation',
            # 'Node',
        ],
        'UI': [
            'Button',
            'Draggable',
            'Tooltip',
            'Slider',
            'ThinSlider',
            'TextField',
            # 'Cursor',
            'InputField',
            'ContentTypes',
            'Checkbox',
            'ButtonList',
            'ButtonGroup',
            'WindowPanel',
            # 'Space',
            'FileBrowser',
            # 'FileButton',
            'FileBrowserSave',
            'DropdownMenu',
            # 'DropdownMenuButton',
            'RadialMenu',
            # 'RadialMenuButton',
            'HealthBar',
        ],
        'Editor': [
            'HotReloader',
            'GridEditor',
            'PixelEditor',
            'ASCIIEditor',
            'ColorPicker',
        ],
        'Scripts': [
            'grid_layout',
            'duplicate',
            'SmoothFollow',
            'Scrollable',
            'NoclipMode',
            'NoclipMode2d',
            'build',
        ],
        'Assets': [
            'models',
            'textures',
        ],
        'Shaders': [
            'unlit_shader',
            'lit_with_shadows_shader',
            'matcap_shader',
            'colored_lights_shader',
            'fresnel_shader',
            'projector_shader',
            'texture_blend_shader',
            'instancing_shader',
            'triplanar_shader',
            'normals_shader',
            'transition_shader',
            'fxaa',
            'ssao',
            'camera_outline_shader',
            'pixelation_shader',
            'camera_contrast',
            'camera_vertical_blur',
            'camera_grayscale',
        ],
        }


    modules = [ursina, ] + get_all_modules_in_package(ursina)

    module_infos = dict()
    for mod in modules:
        with open(mod.__file__) as f:
            source_code = f.read()

        # print(mod.__file__)
        module_base_name = mod.__name__
        if '.' in module_base_name:
            module_base_name = module_base_name.split('.')[-1]

        relative_path = Path(mod.__file__.split('ursina')[-1])

        modinf = analyze_module(source_code, relative_path)
        for cls in modinf.classes:
            cls.module = modinf
        module_infos[module_base_name] = modinf
        module_infos[module_base_name].name = module_base_name
        module_infos[module_base_name].path = Path(mod.__file__)


    # for e in module_infos:
    #     print('---', e)
    # package_info = {mod.__name__ : analyze_module(mod) for mod in modules}
    # module_infos = {key : value for key, value in package_info.items() if not key.startswith('_')}
    class_infos = dict()

    for _module_name, modinf in module_infos.items():
        for clsinf in modinf.classes:
            if clsinf.name.startswith('_'):
                continue
            class_infos[clsinf.name] = clsinf
            # print(clsinf.name)

    def render_url_section(path: Path):
        package_folder = application.package_folder.parent
        if package_folder.name not in path.parts:
            return None

        relative_path = path.relative_to(package_folder)
        github_link = f'https://github.com/pokepetter/{package_folder.name}/tree/{version_tag}/{relative_path.as_posix()}'
        dotted_path = '.'.join(relative_path.with_suffix('').parts)
        return f'<gray>&lt;/&gt;</gray><a href="{github_link}"> {dotted_path}</a>\n'


    def add_links(function_info):
        text = '<div class="links">'

        package_folder = application.package_folder.parent
        path = None
        if function_info.module:
            path = function_info.module.path

        if path is None:
            print_warning('missing url for:', function_info.name)
        else:
            if package_folder.name not in path.parts:
                return None

            relative_path = path.relative_to(package_folder)
            github_link = f'https://github.com/pokepetter/{package_folder.name}/tree/{version_tag}/{relative_path.as_posix()}#L{function_info.line_number}'
            text += f'  <a href="{github_link}" alt="Source Code" title="View Source Code" style="font-size:.75em;">&lt;/&gt;</a>'


        permalink = f'#{function_info.name}'
        text += f'  <a href="{permalink}" onclick="navigator.clipboard.writeText(\'{permalink}\')" title="Copy permalink">◃-</a>'
        text += '</div>'
        return text


    def render_function_info(function_info):
        text = f'{function_info.name}('
        for i, var in enumerate(function_info.input_args):
            if i > 0:
                text += ', '
            if len(function_info.input_args) > 8:
                text += '\n    '

            type_hint = f': {var.type_hint}' if var.type_hint is not None else ''
            if var.default_value != ...:
                default_value = var.default_value if len(str(var.default_value)) < 32 else f'{var.default_value[:29]}...'

            text += f'{var.name}{type_hint}={default_value}'

        if len(function_info.input_args) > 8:
            text += '\n    '
        text += ')'

        return text


    def render_function_section(function_info):
        # if method_info.decorators:
        #     print('decorators:', method_info.name, method_info.decorators)
        function_usage = render_function_info(function_info)

        text = dedent(f'''\
            # {function_info.name}()
            ''')

        if function_usage != f'{function_info.name}()':
            text += f'```\n{function_usage}\n```\n'

        if function_info.comment:
            text += f'<purple>{function_info.comment}</purple>\n'


        # examples_folder = application.package_folder.parent / 'examples'
        example_file_base_name = function_info.name
        if function_info.from_class:
            example_file_base_name = f'{function_info.from_class.name}_{example_file_base_name}'

        for separate_example_file in Path('.').glob(f'{example_file_base_name}_example*.py'):
            print('found external example:', separate_example_file)
            example_name = separate_example_file.stem.split('example')[1]
            if not example_name:
                example_name = 'Example'
            text += f'{example_name}\n'

            with separate_example_file.open('r') as f:
                text += dedent(f'''\
```
{f.read()}
```
''')
        text += add_links(function_info) +'\n'
        text += '<hr></hr>\n'
        return text


    def render_variable_info(var:VarInfo, prefix='.', truncate_at:int=None, add_comment=False):
        type_hint = f': {var.type_hint}' if var.type_hint is not None else ''
        default_value = var.default_value
        if truncate_at is not None and len(str(var.default_value)) > truncate_at:
            default_value = f'{var.default_value[:20]}...'

        return f'{prefix}{var.name} {type_hint}<gray>= {default_value}</gray>'


    def render_variable_section(var):
        text = ''
        text += dedent(f'''\
            # .{var.name}
            ''')

        if var.type_hint is not None:
            text += dedent(f'''\
                'type: `{var.type_hint}`'
                ''')

        text += dedent(f'''\
            <gray>default: </gray>`{var.default_value if len(var.default_value) <= 50 else var.default_value[:50]+'...'}`
            ''')
        if var.comment:
            text += f'<purple>{var.comment}</purple>\n'

        text += dedent('''\
            <hr></hr>
        ''')
        return text


    def render_property_section(method_info):
                text = ''
                text += dedent(f'''\
                    # {method_info.base_name}
                    ''')
                if method_info.comment:
                    text += dedent(f'''\
                        <purple>{method_info.comment}</purple>
                        ''')
                text += add_links(method_info)
                text += dedent('''\
                    <hr></hr>
                    ''')
                return text


    singleton_infos = {key : value for key, value in module_infos.items() if value.is_singleton}
    module_infos = {key : value for key, value in module_infos.items() if not value.is_singleton}

    pages = flatten_list(groups.values())

    # get version tag
    package_version = get_package_version(application.package_folder.parent)
    version_tag = 'v' + package_version.replace('.','_')
    print('package version:', package_version, 'version_tag:', version_tag)
    output_folder = Path(f'api_reference_{version_tag}/')
    output_folder.mkdir(parents=True, exist_ok=True)

    # add front page
    front_page_content = dedent(f'''\
        #center
        ### ursina API Reference
        `{package_version}`
        #left

        ''')

    front_page_content += dedent('''\
        <div class="grid-container">''')

    for group_name, group in groups.items():
        front_page_content += '<div class="group-box">'
        front_page_content += f'<b>{group_name}</b>\n'
        for page_name in group:
            front_page_content += f'<a href="{camel_to_snake(page_name)}.html">{page_name}</a>\n'
        front_page_content += '</div>'

    front_page_content += '</div>'

    html_content = sswg.sswg_to_html(front_page_content, title='ursina API reference')
    html_content = '''
        <style>
        .grid-container {display: grid; grid-template-columns: repeat(auto-fill, minmax(19%, 1fr)); gap: 1%;}
        .group-box {padding: 10px; padding-top: 2em;}
        //.group-box h3 {margin-top:0;}
        </style>''' + html_content
    html_content = html_content.replace('href="sswg.css">', 'href="../sswg.css">')
    html_content = html_content.replace('href="style.css">', 'href="../style.css">')
    with open(output_folder / 'index.html', 'w') as f:
        f.write(html_content)


    # make .sswgs documents for each page
    for name in pages:
        # links = '\n     '.join([f'<a href="#{e}">{e}</a><br>' for e in group])
        # print('look for:', name, name in module_infos)
        if name not in module_infos and name not in class_infos and name not in singleton_infos:
            print_warning('missing module/class info for:', name)
            continue

        text = ''
        text += f'### {name}\n'

        ### modules ###
        if name in module_infos:
            module_info = module_infos[name]
            text += render_url_section(module_info.path)
            module_info.variables = [var for var in module_info.variables if not var.name.startswith('_')]

            if Path(f'images/{name}.webp').exists():
                text += f'#image ../images/{name}.webp\n'

            if module_info.variables:
                text += '## Variables\n'
                for var in module_info.variables:
                    # text += render_variable_info(var, truncate_at=10) + '\n'
                    text += render_variable_section(var)

            if module_info.functions:
                text += '## Functions\n'
                for function_info in [m for m in module_info.functions if not m.name.startswith('_')]:
                    text += render_function_section(function_info)

            if module_info.examples:
                text += '## Examples\n'
                for example in module_info.examples:
                    text += dedent(f'''\
{example.name}
```
{example.content}
```
''')
            for separate_example_file in Path('.').glob(f'{module_info.name}_example*.py'):
                print('found external example:', separate_example_file)
                with separate_example_file.open('r') as f:
                    text += dedent(f'''\
{separate_example_file.stem}
```
{f.read()}
```
''')


        ### classes ###
        elif name in class_infos or name in singleton_infos:
            is_singleton = name in singleton_infos
            if is_singleton:
                name = snake_to_camel(name).title()

            class_info = class_infos[name]

            if is_singleton:
                name = class_info.module.name

            text += render_url_section(class_info.module.path)

            base_classes_in_repo = [bcls for bcls in class_info.base_classes if bcls in pages]
            if base_classes_in_repo and not is_singleton:
                text += '<gray>Inherits</gray> '
                for bcls in base_classes_in_repo:
                    text += f'<a href="{camel_to_snake(bcls)}.html">{bcls}</a>'
                text += '\n'

            if Path(f'images/{name}.webp').exists():
                text += f'#image ../images/{name}.webp\n'

            class_variables= [e for e in class_info.class_variables if not e.name.startswith('_')] # ignore functions starting with underscore (_)
            if class_variables:
                text += '\n## Class Variables\n'
                for var in class_variables:
                    text += render_variable_section(var)

            ### __init__###
            init_methods = [m for m in class_info.methods if m.name == '__init__']
            init_method = init_methods[0] if init_methods else None
            if init_method:
                if not is_singleton:
                    text += dedent(f'''\

## Initialization
```
{class_info.name}{render_function_info(init_method).replace('__init__', '')}
```
                    ''')

                ### attributes (.self) ###
                public_attributes = [e for e in class_info.instance_variables if not e.name.startswith('_')]

                # remove cases with self.some_var = some_var
                public_attributes = [e for e in public_attributes if e.default_value != e.name or e.comment]

                if public_attributes:
                    text += dedent('''\

                        ## Attributes
                    ''')
                for var in public_attributes:
                    text += render_variable_section(var)

            # for m in class_info.methods:
            #     print(m.decorators)

            getters = [m for m in class_info.methods if m.name.endswith('_getter') or 'property' in m.decorators]
            setters = [m for m in class_info.methods if m.name.endswith('_setter') or f'{m.name}.setter' in m.decorators]

            # remove '_getter' and '_setter' from properties made ursina the property generator
            for e in getters:
                if e.name.endswith('_getter'):
                    e.base_name = e.name[:-len('_getter')]
                else:
                    e.base_name = e.name

            for e in setters:
                if e.name.endswith('_setter'):
                    e.base_name = e.name[:-len('_setter')]
                else:
                    e.base_name = e.name

            getters_only = [m for m in getters if m.base_name not in [p.base_name for p in setters]]
            two_way_properties = [m for m in setters if m.base_name not in [p.base_name for p in getters_only]]



            if getters_only:
                text += '\n## Getters\n'
            for method_info in getters_only:
                text += render_property_section(method_info)

            if two_way_properties:
                text += dedent('''\


                    ## Properties
                    ''')
            for method_info in two_way_properties:
                text += render_property_section(method_info)



            public_methods = [m for m in class_info.methods if not m.name.startswith('_') and m not in getters and m not in setters]
            if public_methods:
                text += '## Methods\n'
                for method_info in public_methods:
                    text += render_function_section(method_info) + '\n'

            if class_info.examples:
                # print('EXAMPLE::::')
                text += '## Examples\n'
                for example in class_info.examples:
                    text += dedent(f'''\
{example.name}
```
{example.content}
```
# ''')
            base_name = class_info.name
            if class_info.module.is_singleton:
                base_name = camel_to_snake(base_name)

            external_examples = list(Path('.').glob(f'{base_name}_example*.py'))
            if len(external_examples) > 0 and not class_info.examples:
                text += '## Examples\n'

            for separate_example_file in external_examples:
                example_name = separate_example_file.stem.split('example')[1]
                example_name = example_name.lstrip('_')
                print('found external example:', separate_example_file, example_name)
                if not example_name:
                    example_name = 'Example'
                text += f'{example_name}\n'

                with separate_example_file.open('r') as f:
                    text += f'''\
```
{f.read()}
```
'''

        # # convert sswg to html without having to save it to disk first
        # path = output_folder / f'{camel_to_snake(name)}.sswg'
        # with path.open('w') as f:
        #     f.write(text)
        #     print('saved:', path)

        # add sidebar
        sidebar_content = '<div class="sidebar">'
        sidebar_content += f'`{package_version}`'
        for group_name, group in groups.items():
            sidebar_content += f'\n{group_name}\n'
            for page_name in group:
                sidebar_content += f'  <a href="{camel_to_snake(page_name)}.html">{page_name}</a>\n'
        sidebar_content += '</div>'
        # with open('api_reference_sidebar.sswg', 'w') as f:
        #     f.write(sidebar_content)
        text += sidebar_content

        html_content = sswg.sswg_to_html(text, title=name)
        html_content = html_content.replace('href="sswg.css">', 'href="../sswg.css">')
        html_content = html_content.replace('href="style.css">', 'href="../style.css">')
        html_content += dedent('''\
            <style>
            h1 {margin-bottom:0em; margin-top:1em}
            h2 {margin-bottom:.5em;}
            h3 {margin:0em; font-size:1em;}
            code_block {margin-top:.0em; margin-bottom:1em;}
            hr {border-color:gray; opacity:.25;}
            .links {width:fit-content;}
            .links a {text-decoration:none; opacity:.5;}
            </style>
            ''')

        path = output_folder / f'{camel_to_snake(name)}.html'
        with path.open('w') as f:
            f.write(html_content)

        # print('saved:', path)


    # samples
    # generate sample pages:
    samples_folder = application.package_folder.parent / 'samples'
    generated_sample_pages = dict()

    for sample_name in ('Tic Tac Toe', 'Inventory', 'Pong', 'Minecraft Clone', "Rubik's Cube", 'Clicker Game', 'Platformer', 'FPS', 'Column Graph'):
        file_name = sample_name.replace(' ','_').replace('\'','').lower() + '.py'
        file_path = samples_folder / file_name
        image_path = Path(f'icons/{file_path.stem}.jpg')
        image_code = ''
        image_url = f'icons/{file_path.stem}.jpg'
        github_link = f'https://github.com/pokepetter/ursina/tree/v{package_version}/samples/{file_name}'
        if image_path.exists():
            image_code = f'#image {image_url}'
        else:
            image_code = '#image icons/installation_guide.jpg'
            print('no image:', image_path)

        if file_path.exists():
            # print(f'generate webpage for: {sample_name} -> {file_path.stem}.html')
            with file_path.open() as source_file:
                source_code = source_file.read()

            with open(f'{file_path.stem}.html', 'w') as file:
                file.write(sswg.sswg_to_html(
                        dedent(f'''\
                        #title ursina engine documentation
                        #insert menu.sswg

                        ### {sample_name}
                        <a href="{github_link}">{github_link}</a>

                        {image_code}
                        ```\n''')
                        + source_code
                        + '\n```'
                    )
                )
                generated_sample_pages[sample_name] = (f'{file_path.stem}.html', image_url)
        else:
            print('sample not found:', file_path)

    samples_page_content = dedent('''\
        #title ursina engine samples
        #insert menu.sswg

        ### Samples
        ## Single File
        ''')
    for name, urls in generated_sample_pages.items():
        html_url, image_url = urls
        samples_page_content += f'[{name}, {html_url}, {image_url}]\n'

    samples_page_content += dedent('''\
        ## Projects
        [Value of Life, https://github.com/pokepetter/ld44_life_is_currency, icons/value_of_life_icon.jpg]
        [Castaway, https://github.com/pokepetter/pyweek_30_castaway, icons/castaway_icon.jpg]
        [Protein Visualization, https://github.com/HarrisonTCodes/ursina-proteins, icons/protein_visualization_icon.jpg]
        ''')
    with open('samples.sswg', 'w') as f:
        f.write(samples_page_content)


if __name__ == '__main__':
    generate_documentation()


from ursina import Func, Shader, camera, color, window

outline_shader = Shader(fragment='''
#version 430

uniform sampler2D tex;
uniform sampler2D dtex;
uniform vec2 window_size;
uniform float far = 100.;
uniform float near = 1.;
uniform vec4 outline_color;

in vec2 uv;
out vec4 out_color;

float eye_depth(float depth) {
    return far * near / ((near - far) * depth + far);
}

void main() {
    vec4 color = texture(tex, uv).rgba;
    float depth = eye_depth(texture(dtex, uv).r);

    vec2 pixel_size = vec2(1./window_size.x, 1./window_size.y);
    // vec2 pixel_size = vec2(.001);

    float n = eye_depth(texture(dtex, uv + pixel_size * vec2(0, 1)).r);
    float e = eye_depth(texture(dtex, uv + pixel_size * vec2(1, 0)).r);
    float s = eye_depth(texture(dtex, uv + pixel_size * vec2(0, -1)).r);
    float w = eye_depth(texture(dtex, uv + pixel_size * vec2(-1, 0)).r);

    if (n - s > 0.05 || w - e > 0.05 || e - w > 0.05 || s - n > 0.05)
        color = outline_color;

    out_color = color;
}
''',
default_input={
    'window_size': window.size,
    'far': Func(getattr, camera, 'clip_plane_far'),
    'near': .1,

    'outline_color': color.black,

}
)



if __name__ == '__main__':
    from ursina import *

    app = Ursina()
    e = Entity(model='sphere', color=color.white)
    e = Entity(model='cube', y=-1)
    Entity(model='plane', scale=100, y=-10)
    camera.shader = outline_shader
    camera.far_clip_plane = 1000
    camera.set_shader_input('near', .01)

    def input(key):
        if key == 'space':
            if not camera.shader:
                camera.shader = outline_shader
            else:
                camera.shader = None

    EditorCamera()

    app.run()



from ursina import *

from ursina.shaders import basic_lighting_shader, colored_lights_shader

app = Ursina()

Entity(model='procedural_rock_0', shader=colored_lights_shader)
EditorCamera()

app.run()



from ursina import Shader, Vec2, Vec3

vertex_animation_shader = Shader(
vertex='''
#version 330 core

uniform mat4 p3d_ModelViewProjectionMatrix;
in vec4 p3d_Vertex;
in vec2 p3d_MultiTexCoord0;
out vec2 uvs;
uniform vec2 texture_scale;
uniform vec2 texture_offset;

in vec4 p3d_Color;

uniform sampler2D frame_texture; // texture containing all animation frames
uniform float frame_index;
uniform float total_frames;
uniform vec3 pos_min;
uniform vec3 pos_max;
uniform float num_verts;

out vec4 vertex_color;

void main() {
    float frame_height = 1.0 / total_frames; // Each frame's height in the texture
    vec3 pos_offset = texture(frame_texture, vec2((gl_VertexID/num_verts) + ((1./num_verts)*.5), (frame_index/total_frames) + (frame_height*.5))).rgb;
    vec3 denormalized_pos_offset = mix(pos_min, pos_max, pos_offset.xyz);
    gl_Position = p3d_ModelViewProjectionMatrix * vec4(denormalized_pos_offset, 1.0);
    uvs = (p3d_MultiTexCoord0 * texture_scale) + texture_offset;
    vertex_color = p3d_Color;
    vertex_color = texture(frame_texture, vec2((gl_VertexID/num_verts) + ((1./num_verts)*.5) +.5, (frame_index/total_frames) + (frame_height*.5)));
}

''',

fragment='''
#version 330 core

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;
in vec2 uvs;
out vec4 fragColor;

in vec4 vertex_color;

void main() {
    vec4 color = texture(p3d_Texture0, uvs) * p3d_ColorScale * vertex_color;
    fragColor = color.rgba;
}

''',
default_input = {
    'texture_scale' : Vec2(1,1),
    'texture_offset' : Vec2(0.0, 0.0),

    'frame_index': 0,
}
)


if __name__ == '__main__':
    from ursina import EditorCamera, Entity, Ursina, load_texture
    app = Ursina()
    e = Entity(model='cube', shader=vertex_animation_shader, texture='grass')
    animation_texture = load_texture('grass')
    e.set_shader_input('total_frames', animation_texture.height)
    e.set_shader_input('num_verts', animation_texture.width)
    e.set_shader_input('frame_texture', animation_texture)

    e.set_shader_input('pos_min', Vec3(-.5,-.5,-.5))
    e.set_shader_input('pos_max', Vec3(.5,.5,.5))
    EditorCamera()

    e.animate_shader_input('frame_index', 10, duration=3, loop=True)

    app.run()


from ursina import *





app = Ursina()


# Entity(model='cube', scale=(16,.1,16))

from ursina.prefabs.primitives import *

camera.far = 100

from panda3d.core import *
from direct.showbase.ShowBase import ShowBase
from direct.actor.Actor import Actor
from direct.gui.DirectGui import *

#place the camera
base.trackball.node().setPos(0, 15, 1)
base.trackball.node().setHpr(0, 40, 0)

#the shaders...
#vertex shader:
v_shader='''#version 140

            struct p3d_LightSourceParameters {
              vec4 color;
              vec3 spotDirection;
              sampler2DShadow shadowMap;
              mat4 shadowViewMatrix;

            };

            uniform p3d_LightSourceParameters my_light;
            uniform mat4 p3d_ModelViewProjectionMatrix;
            uniform mat3 p3d_NormalMatrix;
            uniform mat4 p3d_ModelViewMatrix;

            in vec4 p3d_Vertex;
            in vec3 p3d_Normal;
            in vec2 p3d_MultiTexCoord0;

            out vec2 uv;
            out vec4 shadow_uv;
            out vec3 normal;

            void main()
                {
                //position
                gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex;
                //normal
                normal = p3d_NormalMatrix * p3d_Normal;
                //uv
                uv = p3d_MultiTexCoord0;
                //shadows
                shadow_uv = my_light.shadowViewMatrix * (p3d_ModelViewMatrix * p3d_Vertex);
                }'''
#fragment shader
f_shader='''#version 140

        struct p3d_LightSourceParameters {
          vec4 color;
          vec3 spotDirection;
          sampler2DShadow shadowMap;
          mat4 shadowMatrix;
        };

        uniform p3d_LightSourceParameters my_light;
        uniform sampler2D p3d_Texture0;
        uniform vec3 camera_pos;
        uniform float shadow_blur;

        in vec2 uv;
        in vec4 shadow_uv;
        in vec3 normal;

        out vec4 color;

        float textureProjSoft(sampler2DShadow tex, vec4 uv, float bias, float blur)
            {
            float result = textureProj(tex, uv, bias);
            result += textureProj(tex, vec4(uv.xy + vec2( -0.326212, -0.405805)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(-0.840144, -0.073580)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(-0.695914, 0.457137)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(-0.203345, 0.620716)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.962340, -0.194983)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.473434, -0.480026)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.519456, 0.767022)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.185461, -0.893124)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.507431, 0.064425)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(0.896420, 0.412458)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(-0.321940, -0.932615)*blur, uv.z-bias, uv.w));
            result += textureProj(tex, vec4(uv.xy + vec2(-0.791559, -0.597705)*blur, uv.z-bias, uv.w));
            return result/13.0;
            }

        void main()
            {
            //base color
            vec3 ambient=vec3(0.1, 0.1, 0.2);
            //texture
            vec4 tex=texture(p3d_Texture0, uv);
            //light ..sort of, not important
            vec3 light=vec3(0,0,0);

            //shadows
            //float shadow= textureProj(my_light.shadowMap,shadow_uv); //meh :|
            float shadow= textureProjSoft(my_light.shadowMap, shadow_uv, 0.0001, shadow_blur);//yay! :)

            //make the shadow brighter
            shadow=0.5+shadow*0.5;

            color=vec4(tex.rgb*(light*shadow+ambient), tex.a);

            }'''
shader = Shader.make(Shader.SL_GLSL,v_shader, f_shader)

#make some floor
cm = CardMaker('')
cm.set_frame(-10, 10, -10, 10)
floor=render.attach_new_node(cm.generate())
floor.set_p(-90)
#set a texture
floor.set_texture(loader.load_texture('maps/grid.rgb'))
floor.set_shader(shader)

#load some model
panda=Actor('panda-model', {'walk': 'panda-walk4'})
panda.reparent_to(render)
panda.set_scale(0.005)
panda.loop('walk')
panda.set_shader(shader)

#light
my_light = render.attach_new_node(DirectionalLight("Spot"))
my_light.node().set_shadow_caster(True, 512, 512)
my_light.node().set_color((0.9, 0.9, 0.8, 1.0))
#my_light.node().showFrustum()
my_light.node().get_lens().set_fov(40)
my_light.node().get_lens().set_near_far(0.1, 30)
render.setLight(my_light)
my_light.set_pos(-20, 0, 20)
my_light.look_at(0, 0, 0)
render.set_shader_input('my_light',my_light)
render.set_shader_input('shadow_blur',0.2)


def set_softness():
    v=float(slider['value'])
    render.set_shader_input('shadow_blur', v)

slider = DirectSlider(range=(0, 0.5), value=0.2, scale=0.5, pos=(-0.8,0.0,0.9), command=set_softness)


EditorCamera()

RedCube(x=2, shader=shader)
GreenSphere(y=2, shader=shader)

app.run()



from ursina import *



voxel_tool = '''
from ursina import *
from ursina.prefabs.first_person_controller import FirstPersonController


app = Ursina()

class Voxel(Button):
    def __init__(self, position=(0,0,0)):
        super().__init__(
            parent = scene,
            position = position,
            model = 'cube',
            origin_y = .5,
            texture = 'white_cube',
            color = color.color(0, 0, random.uniform(.9, 1.0)),
            highlight_color = color.lime,
        )


    def input(self, key):
        if self.hovered:
            if key == 'left mouse down':
                voxel = Voxel(position=self.position + mouse.normal)

            if key == 'right mouse down':
                destroy(self)


for z in range(8):
    for x in range(8):
        voxel = Voxel(position=(x,0,z))


player = FirstPersonController()
app.run()
'''
window.color = color.color(220, .23, .15)
tf = TextField()
tf.text_entity.text_colors['default'] = color.color(219, .0, .95)
tf.text_entity.text_colors['class_color'] = color.color(40, .61, .9)
tf.text_entity.text_colors['kw_color'] = color.color(210, .59, .94)
tf.text_entity.text_colors['func_color'] = color.color(250, .46, .87)
tf.text_entity.text_colors['param_clor'] = color.color(30, .71, .92)
tf.text_entity.text_colors['string_color'] = color.color(90, .48, .86)

# tf.text_entity.text_colors['default'] = color._200
# tf.text_entity.text_colors['class_color'] = color._250
# tf.text_entity.text_colors['kw_color'] = color._250
# tf.text_entity.text_colors['func_color'] = color._250
# tf.text_entity.text_colors['param_clor'] = color._250
# tf.text_entity.text_colors['string_color'] = color._250

tf.replacements = {

    'from ':    f'☾kw_color☽from ☾default☽',
    'import ':  f'☾kw_color☽import ☾default☽',
    'def ':     f'☾kw_color☽def ☾default☽',
    'for ':     f'☾kw_color☽for ☾default☽',
    'if ':      f'☾kw_color☽if ☾default☽',
    ' in ':     f'☾kw_color☽ in ☾default☽',

    'print(':   f'☾func_color☽print☾default☽(',
    'range(':   f'☾func_color☽range☾default☽(',
    '__init__': f'☾func_color☽__init__☾default☽',
    'super':    f'☾func_color☽super☾default☽',

    'class ':   f'☾class_color☽class ☾default☽',
    'Entity':   f'☾lime☽Entity☾default☽',
    'self.':    f'☾class_color☽self☾default☽.',
    '(self)':   f'(☾class_color☽self☾default☽)',
    'self,':    f'☾class_color☽self☾default☽,',

    'highlight_color = ':    f'☾param_clor☽highlight_color☾default☽ = ',

    '\',':    f'\',☾default☽',   # end quote
    '\':':    f'\':☾default☽',   # end quote
    '\')':    f'\')☾default☽',   # end quote
    '\'':    f'☾string_color☽\'', # start quote
    }
for name in tf.attributes:
    tf.replacements[f'{name}='] = f'☾param_clor☽{name}☾default☽='
    tf.replacements[f'{name} = '] = f'☾param_clor☽{name}☾default☽ = '

# tf.line_numbers.enabled = True
tf.cursor.enabled = False
tf.add_text(voxel_tool[1:])
tf.text_entity.text = multireplace(tf.text, tf.replacements)
# tf.line_numbers.text = '\n'.join([str(e) for e in range(len(tf.text.split('\n')))])
# tf.line_numbers.color = color.gray
tf.text_entity.enabled = False
tf.bg.enabled = False
# tf.text_entity.appear(speed=.004)



from ursina import *

import numpy as np
import mcubes

app = Ursina()
# voxels = list()

# print(blank)

slice = '''
000000000000000000000
011111110000000000000
011111110000000000000
011001111111111111110
011001111111111111110
011001111111111111110
011111110000000000000
011111110000000000000
000000000000000000000
'''
# Entity(model='cube', texture='pot')
t = load_texture('pot')
slice = ''
for row in t.pixels:
    new_row = list()

    for p in row:
        col = color.rgba(*p)
        if col == color.white:
            slice += '0'
        else:
            slice += '1'

    slice += '\n'


print(slice)
slice = [[bool(i=='1') for i in e] for e in slice.strip().split('\n')]

blank = (('0'*len(slice[0])) + '\n') * len(slice)
print(blank)
blank = [[bool(i=='1') for i in e] for e in blank.strip().split('\n')]

voxels = (blank, slice, slice, slice, blank)

# Create a data volumecmd
X, Y, Z = np.mgrid[:len(slice[0]), :len(slice), :4]
# u = (X-15)**2 + (Y-15)**2 + (Z-15)**2 - 8**2
u = np.array(voxels)
# print(u)

# Extract the 0-isosurface
vertices, triangles = mcubes.marching_cubes(u, 0)
vertices, triangles = vertices.tolist(), triangles.tolist()
# print(vertices, triangles)

e = Entity(model=Mesh(vertices, triangles, mode='triangle'))
e.model.colorize()
for c in e.model.colors:
    if c != c:
        print('lol')
    print(c)
# e.reflectivity = 1
# e.flip_faces()
e.scale = .1
e.scale_y *= -1
EditorCamera()
app.run()
# # Export the result to sphere.dae
# mcubes.export_mesh(vertices, triangles, "sphere.dae", "MySphere")



import sys
sys.path.append('..')
from ursina import *


class PuzzleGame(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'puzzle_game'

        player = FirstPersonController()
        player.parent = self

    def load_level(self, name):
        from PIL import Image
        im = Image.open(application.internal_textures_folder + name + '.png')
        pix = im.load()
        print (im.size)
        for y in range(im.size[1]):
            for x in range(im.size[0]):
                print('color:', pix[x,y])

    def input(self, key):
        if key == 'l':
            self.load_level('white_cube')



class Voxel(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'voxel'
        self.model = 'cube'
        self.origin = (0, .5, 0)
        self.scale_y = 8
        self.collider = 'box'
        self.texture = 'white_cube'
        self.color = color.color(0, 0, random.uniform(.9, 1.0))


    def input(self, key):
        if self.hovered:
            if key == 'left mouse down' and self.y < 7:
                self.y += 1

            if key == 'right mouse down' and self.y > 0:
                self.y -= 1


class FirstPersonController(Entity):

    def __init__(self):
        super().__init__()
        self.speed = .1

        cursor = Panel()
        cursor.color = color.dark_gray
        cursor.scale *= .01
        cursor.rotation_z = 45


    def start(self):
        self.position = (0, 2, 1)
        camera.parent = self
        camera.position = (0,0,0)
        camera.rotation = (0,0,0)
        camera.fov = 90
        mouse.locked = True


    def update(self):


        self.position += self.right * held_keys['d'] * self.speed
        self.position += self.forward * held_keys['w'] * self.speed
        self.position += self.left * held_keys['a'] * self.speed
        self.position += self.back * held_keys['s'] * self.speed

        self.rotation_y += mouse.velocity[0] * 20
        camera.rotation_x -= mouse.velocity[1] * 20
        camera.rotation_x = clamp(camera.rotation_x, -90, 90)


if __name__ == '__main__':
    app = main.Ursina()
    scene = PuzzleGame()
    app.run()



from ursina import *


class ToggleSideview():

    def on_click(self):
        camera_original_parent = camera.parent
        camera.reparent_to(scene.editor.editor_camera.camera_pivot)
        scene.editor.editor_camera.camera_pivot.rotation = (0,0,0)
        camera.reparent_to(camera_original_parent)



from ursina import *; ssao_shader = Shader(fragment=''' #version 430
uniform sampler2D tex;
uniform sampler2D dtex;
uniform sampler2D ntex;
in vec2 uv;

uniform sampler2D noise_texture;
in flat int numsamples;
in float radius;
in float amount;
in float strength;
in float falloff;

uniform vec4 params1;
uniform vec4 params2;

out vec4 color;


vec3 sphere[16] = vec3[](
    vec3(0.53812504, 0.18565957, -0.43192).xyz,
    vec3(0.13790712, 0.24864247, 0.44301823).xyz,
    vec3(0.33715037, 0.56794053, -0.005789503).xyz,
    vec3(-0.6999805, -0.04511441, -0.0019965635).xyz,
    vec3(0.06896307, -0.15983082, -0.85477847).xyz,
    vec3(0.056099437, 0.006954967, -0.1843352).xyz,
    vec3(-0.014653638, 0.14027752, 0.0762037).xyz,
    vec3(0.010019933, -0.1924225, -0.034443386).xyz,
    vec3(-0.35775623, -0.5301969, -0.43581226).xyz,
    vec3(-0.3169221, 0.106360726, 0.015860917).xyz,
    vec3(0.010350345, -0.58698344, 0.0046293875).xyz,
    vec3(-0.08972908, -0.49408212, 0.3287904).xyz,
    vec3(0.7119986, -0.0154690035, -0.09183723).xyz,
    vec3(-0.053382345, 0.059675813, -0.5411899).xyz,
    vec3(0.035267662, -0.063188605, 0.54602677).xyz,
    vec3(-0.47761092, 0.2847911, -0.0271716).xyz
);


void main() {
    // vec3 rgb = normalize(texture(ntex, uv).rgb);
    //color.rgb = (1.-rgb)*10;

    float pixel_depth = texture(dtex, uv).a;
    vec3 pixel_normal = (texture(ntex, uv).xyz * 2.0 - 1.0);
    vec3 random_vector = normalize((texture(noise_texture, uv * 18.0 + pixel_depth + pixel_normal.xy).xyz * 2.0) - vec3(1.0)).xyz;
    float occlusion = 0.0;
    float radius = params1.z / pixel_depth;
    // float radius = radius / pixel_depth;
    float depth_difference;
    vec3 sample_normal;
    vec3 ray;
    // for(int i = 0; i < numsamples; ++i) {
    //     ray = radius * reflect(sphere[i], random_vector);
    //     sample_normal = (texture(ntex, uv + ray.xy).xyz * 2.0 - 1.0);
    //     depth_difference = (pixel_depth - texture(dtex, uv + ray.xy).r);
    //     // occlusion += step(params2.y, depth_difference) * (1.0 - dot(sample_normal.xyz, pixel_normal)) * (1.0 - smoothstep(params2.y, params2.x, depth_difference));
    //     // occlusion += step(params2.y, depth_difference) * (1.0 - dot(sample_normal.xyz, pixel_normal)) * (1.0 - smoothstep(params2.y, params2.x, depth_difference));
    //     occlusion += 100.;
    //     // occlusion += step(falloff, depth_difference) * (1.0 - dot(sample_normal.xyz, pixel_normal)) * (1.0 - smoothstep(falloff, strength, depth_difference));
    // }
    ray = radius * reflect(sphere[0], random_vector);
    sample_normal = (texture(ntex, uv + ray.xy).xyz * 2.0 - 1.0);
    depth_difference = (pixel_depth - texture(dtex, uv + ray.xy).r);
    // occlusion += step(params2.y, depth_difference) * (1.0 - dot(sample_normal.xyz, pixel_normal)) * (1.0 - smoothstep(params2.y, params2.x, depth_difference));
    occlusion += step(params2.y, depth_difference) * (1.0 - dot(sample_normal.xyz, pixel_normal)) * (1.0 - smoothstep(params2.y, params2.x, depth_difference));
    occlusion *= 100.0;

    float amount = -amount / numsamples;
    color.rgb = texture(tex, uv).rgb + (occlusion * params1.y);
    // color.rgb = vec3(1.,1.,1.) + (occlusion * amount);
    // pixel_depth = (1.-pixel_depth) * 10;
    // color.rgb = vec3(1.,1.,1) * random_vector;
    // color.rgb = rgb;
    color.a = 1.;
}
// uniform mat4 p3d_ProjectionMatrixInverse;
// vec4 getViewPos(vec2 texCoord)
// {
// 	float x = texCoord.s * 2.0 - 1.0;
// 	float y = texCoord.t * 2.0 - 1.0;
// 	// Assume we have a normal depth range between 0.0 and 1.0
// 	float z = texture(ntex, texCoord).r * 2.0 - 1.0;
// 	vec4 posProj = vec4(x, y, z, 1.0);
// 	// vec4 posView = u_inverseProjectionMatrix * posProj;
// 	vec4 posView = p3d_ProjectionMatrixInverse * posProj;
//
// 	posView /= posView.w;
//
// 	return posView;
// }
// #define KERNEL_SIZE 16
//
// // uniform vec3 u_kernel[KERNEL_SIZE];
// uniform mat4 p3d_ProjectionMatrix;
// // uniform float u_radius;
//
// #define CAP_MIN_DISTANCE 0.0001 // This constant removes artifacts caused by neighbour fragments with minimal depth difference.
// #define CAP_MAX_DISTANCE 0.005 // This constant avoids the influence of fragments, which are too far away.
//
//
// void main() {
// 	vec4 posView = getViewPos(uv);
// 	vec3 normalView = normalize(texture(ntex, uv).xyz * 2.0 - 1.0);
// 	vec3 randomVector = normalize(texture(noise_texture, uv * 1.).xyz * 2.0 - 1.0);
// 	// vec3 randomVector = normalize(texture(noise_texture, v_texCoord * u_rotationNoiseScale).xyz * 2.0 - 1.0);
// 	vec3 tangentView = normalize(randomVector - dot(randomVector, normalView) * normalView);
// 	vec3 bitangentView = cross(normalView, tangentView);
// 	mat3 kernelMatrix = mat3(tangentView, bitangentView, normalView);
// 	// Go through the kernel samples and create occlusion factor.
// 	float occlusion = 0.0;
//
// 	for (int i = 0; i < KERNEL_SIZE; i++)
// 	{
// 		// vec3 sampleVectorView = kernelMatrix * u_kernel[i];
// 		vec3 sampleVectorView = kernelMatrix * sphere[i];
// 		vec4 samplePointView = posView + radius * vec4(sampleVectorView, 0.0);
// 		// vec4 samplePointNDC = u_projectionMatrix * samplePointView;
// 		vec4 samplePointNDC = p3d_ProjectionMatrix * samplePointView;
// 		samplePointNDC /= samplePointNDC.w;
// 		vec2 samplePointTexCoord = samplePointNDC.xy * 0.5 + 0.5;
// 		float zSceneNDC = texture(dtex, samplePointTexCoord).r * 2.0 - 1.0;
// 		float delta = samplePointNDC.z - zSceneNDC;
// 		// If scene fragment is before (smaller in z) sample point, increase occlusion.
// 		if (delta > CAP_MIN_DISTANCE && delta < CAP_MAX_DISTANCE)
// 		{
// 			occlusion += 1.0;
// 		}
// 	}
//
// 	// No occlusion gets white, full occlusion gets black.
// 	occlusion = 1.0 - occlusion / (float(KERNEL_SIZE) - 1.0);
//
// 	color = vec4(occlusion, occlusion, occlusion, 1.0);
//
// }
''',
default_input = dict(
    numsamples = 64,
    # -float(config.amount) / config.numsamples  #params1.y
    amount = 2.0,
    radius = 0.05,

    strength = 0.01,
    falloff = 0.000002,
)
)
app = Ursina()

camera.clip_plane_far = 100
camera.shader = ssao_shader
camera.set_shader_input('noise_texture', load_texture('noise')._texture)
camera.set_shader_input("params1", (64, -float(2.0) / 64, .05, 0))
camera.set_shader_input("params2", (.01, .000002, 0, 0))

base.render.setShaderAuto()


Entity(model='plane', scale=8)
Entity(model='cube', color=color.red, y=.5)
Entity(model='sphere', color=color.yellow, y=1, x=.75)

EditorCamera()
app.run()


# from direct.filter.CommonFilters import CommonFilters
# from direct.showbase.ShowBase import ShowBase
#
# app = ShowBase()
#
# m = loader.loadModel("panda.egg")
# m.reparentTo(render)
# # m.setY(20)
#
# filters = CommonFilters(base.win, base.cam)
# filters.setAmbientOcclusion(
#     numsamples=64,
#     # radius=0.05,
#     # amount=2.0,
#     # strength=0.01,
#     # falloff= 0.000002,
# )
#
# app.run()



from ursina import Button, Text, Quad


class Checkbox(Button):
    def __init__(self, start_value=False, **kwargs):
        super().__init__(scale=Text.size, model=Quad(radius=.25))
        self.start_value = start_value
        self.value = start_value
        for key, value in kwargs.items():
            setattr(self, key, value)


    def on_click(self):
        self.value = not self.value


    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = new_value
        self.text = ' x'[int(new_value)]


if __name__ == '__main__':
    from ursina import Ursina, Slider
    app = Ursina()
    Checkbox(start_value=True)
    Checkbox(x=.1, start_value=False)
    Slider(y=-.1)
    app.run()



from ursina.editor.level_editor import *


def stretch_model(mesh, scale, limit=.25, scale_multiplier=1, regenerate=False):
    verts = [Vec3(*e) for e in mesh.vertices]
    mesh.uvs = [Vec2(*e) for e in mesh.uvs]

    for i, v in enumerate(verts):
        for j in [0,1,2]:
            if v[j] <= -limit:
                verts[i][j] += .5 + (scale_multiplier/2)
                verts[i][j] -= scale[j] / 2
                if mesh.uvs:
                    mesh.uvs[i][0] += .5 + (scale_multiplier/2)

            elif v[j] >= limit:
                verts[i][j] -= .5
                verts[i][j] += scale[j] / 2

        verts[i] /= scale

    mesh.vertices = verts
    print('----', mesh.vertices)

    # if mesh.uvs:


    if regenerate:
        mesh.generate()




if not load_model('sliceable_cube.ursinamesh', path=Path(__file__).parent):
    m = load_model('sliceable_cube.blend', path=Path(__file__).parent)
    m.save('sliceable_cube.ursinamesh')



@generate_properties_for_class()
class SlicedCube(Entity):
    default_values = Entity.default_values | dict(model=None, shader='lit_with_shadows_shader', texture='white_cube', collider='box', name='sliced_cube', scale_multiplier=1) # combine dicts
    def __init__(self, stretchable_mesh='sliceable_cube', **kwargs):
        kwargs = __class__.default_values | kwargs

        if isinstance(stretchable_mesh, str):
            # Entity(model=stretchable_mesh)
            stretchable_mesh = load_model(stretchable_mesh, use_deepcopy=True)
            print('--------------------', 'asdasdølaksdjkljload:', stretchable_mesh)
        self.stretchable_mesh = stretchable_mesh
        # self.original_vertices = model.vertices
        super().__init__(**__class__.default_values | kwargs)
        self.model = deepcopy(self.stretchable_mesh)
        # self.model = Mesh(vertices=self.stretchable_mesh.vertices, uvs=self.stretchable_mesh.uvs)
        self.model.name = 'cube'
        self.scale_multiplier = kwargs['scale_multiplier']
        self.scale = kwargs['scale']
        self.texture = kwargs['texture']


    def __deepcopy__(self, memo):
        return eval(repr(self))

    # def scale_getter(self):
    #     return super().scale_getter()

    # def scale_setter(self, value):
    #     super().scale_setter(value)
    #     print('------------uuuuuuuuuu')
    #     if self.model:  # ensure init is done and that there's a Mesh as model
    #         self.generate()

    def generate(self):
        print('update model',self.scale)

        self.model.vertices = self.stretchable_mesh.vertices
        self.model.uvs = self.stretchable_mesh.uvs
        stretch_model(self.model, self.world_scale)
        self.model.generate()

    def __setattr__(self, name, value):
        super().__setattr__(name, value)
        if self.model and name in ('scale', 'scale_x', 'scale_y', 'scale_z', 'transform', 'world_transform'):
            self.generate()



if __name__ == '__main__':
    app = Ursina(borderless=False)
    level_editor = LevelEditor()
    level_editor.goto_scene(0,0)

    sliced_cube = SlicedCube(selectable=True, texture='sliceable_cube_template', shader='unlit_shader', scale_multiplier=1.5)
    def input(key):
        if key == 'space':
            sliced_cube.generate()


    level_editor.add_entity(sliced_cube)
    app.run()



from ursina import *

app = Ursina()

cursor_3d = Entity(model='icosphere', scale=.1, color=color.azure)
plane = Entity(model='plane', scale=10, collider='box', color=color.light_gray)
EditorCamera(rotation_x=30)

def update():
    if mouse.world_point:
        cursor_3d.position = mouse.world_point

app.run()


from ursina import color
from ursina.shader import Shader

transition_shader = Shader(name='transition_shader', language=Shader.GLSL, fragment='''
#version 140

uniform sampler2D p3d_Texture0;
uniform vec4 p3d_ColorScale;

in vec2 uv;
out vec4 COLOR;

uniform sampler2D mask_texture;
uniform float min_cutoff;
uniform float max_cutoff;
uniform float smooth_size;

void main() {
    vec3 color = texture2D(p3d_Texture0, uv).rgb * p3d_ColorScale.rgb;
    float value = texture(mask_texture, uv).r;

    float alpha = 1.;
    if (value <= min_cutoff || value > max_cutoff) {
        alpha = 0.;
    }

    COLOR = vec4(color.rgb, alpha * p3d_ColorScale.a);
}

''',

default_input = {
    'min_cutoff' : 0,
    'max_cutoff' : 1,
    'smooth_size' : .25,
}
)

if __name__ == '__main__':
    from ursina import *
    app = Ursina()
    window.color = color._16
    Texture.default_filtering = 'bilinear'

    e = Entity(model='quad', shader=transition_shader, scale=5, cutoff=0,
        texture='shore', color=color.azure
        )
    # mask = load_texture('explosion_particle')
    # mask = load_texture('transition_shader_example_texture')
    # mask = load_texture('perlin_noise')
    mask = load_texture('shore')
    # print(mask)
    e.set_shader_input('mask_texture', mask)
    EditorCamera()

    min_cutoff_slider = Slider(0, 1, dynamic=True, y=-.4)
    def on_value_changed():
        e.set_shader_input('min_cutoff', min_cutoff_slider.value)
    min_cutoff_slider.on_value_changed = on_value_changed

    max_cutoff_slider = Slider(0, 1, default=1, dynamic=True, y=-.45)
    def on_value_changed():
        e.set_shader_input('max_cutoff', max_cutoff_slider.value)
    max_cutoff_slider.on_value_changed = on_value_changed

    # smooth_slider = Slider(0, 1, dynamic=True, y=-.45)
    # def on_value_changed():
    #     e.set_shader_input('smooth_size', smooth_slider.value)
    # smooth_slider.on_value_changed = on_value_changed

    def input(key):
        if key == 'space':
            e.cutoff = 0
            e.animate('cutoff', 1, duration=.1, curve=curve.linear, delay=.05)
        # e.scale=4
        # e.animate('scale', Vec3(7,7,7), duration=.1, curve=curve.linear, delay=.05)

    # Sprite('radial_gradient')

    app.run()



from ursina import *

def rotate_around_point(point, origin, deg):
    from math import pi, cos, sin

    angle_rad = deg/180 * pi
    cos_angle = cos(angle_rad)
    sin_angle = sin(angle_rad)
    dx = point[0] - origin[0]
    dy = point[1] - origin[1]

    return (
        origin[0] + (dx*cos_angle - dy*sin_angle),
        origin[1] + (dx*sin_angle + dy*cos_angle)
        )

class Quad(Mesh):
    def __init__(self, radius=.1, segments=8, aspect=1, scale=(1,1), mode='ngon', **kwargs):


        super().__init__()
        self.vertices = [Vec2(0,0), Vec2(1,0), Vec2(1,1), Vec2(0,1)]
        self.radius = radius
        self.mode = mode

        segments += 1
        if segments > 1:
            offsets = [Vec2(-radius,0), Vec2(0,-radius), Vec2(radius,0), Vec2(0,radius)]
            new_verts = []

            for j, v in enumerate(self.vertices):
                point = v + offsets[j]
                new_verts.append(point)
                for i in range(segments):
                    point = rotate_around_point(point, v, 90/segments)
                    new_verts.append(Vec2(*point))

            self.vertices = new_verts


        self.uvs = list()
        for v in self.vertices:
            self.uvs.append((v[0], v[1]))

        # scale corners horizontally with aspect
        for v in self.vertices:
            if v[0] < .5:
                v[0] /= aspect
            else:
                v[0] = lerp(v[0], 1, 1-(1/aspect))

        # move edges out to keep nice corners
        for v in self.vertices:
            if v[0] > .5:
                v[0] += (scale[0]-1)
            if v[1] > .5:
                v[1] += (scale[1]-1)


        # center mesh
        offset = sum(self.vertices) / len(self.vertices)
        self.vertices = [Vec3(v[0]-offset[0], v[1]-offset[1], 0) for v in self.vertices]


        # make the line connect back to start
        if mode == 'line':
            self.vertices.append(self.vertices[0])

        for key, value in kwargs.items():
            setattr(self, key, value)

        self.generate()
        mesh_importer.imported_meshes[f'Quad(radius={radius}, segments={segments}, aspect={aspect})'] = self




if __name__ == '__main__':
    app = Ursina()
    from time import perf_counter
    t = perf_counter()
    Entity(model=Quad(scale=(3,1), thickness=3, segments=3, mode='line'), color = color.hsv(0,1,1,.7))
    Entity(scale=(3,1), model=Quad(aspect=3), color = color.hsv(60,1,1,.3))
    print('-------', (perf_counter() - t)*100)
    origin = Entity(model='quad', color=color.orange, scale=(.05, .05))
    # ed = EditorCamera(rotation_speed = 200, panning_speed=200)
    camera.z = -5
    app.run()



from ursina import *



app = Ursina()

ground_plane = Entity(model='plane', scale=10, collider='box')

class Node(Draggable):
    def __init__(self, **kwargs):
        super().__init__(
            parent=scene,
            model='sphere',
            # origin_y=-.5,
            # position=mouse.world_point,
            color=color.orange,
            rotation_x=90
        )

        self.temp_arrow = Entity(parent=self, model='cube', scale=.05, origin_z=-.5, color=color.red, x=.1)
        self.connections = list()
        self.connecting = False

        self.arrows = list()


        for key, value in kwargs.items():
            setattr(self, key, value)


    def input(self, key):
        if key == 'c' and self.hovered:
            self.connecting = True

        if self.connecting and key == 'c up' and isinstance(mouse.hovered_entity, Node) and mouse.hovered_entity is not self:
            print('connect')
            self.connecting = False
            self.connections.append(mouse.hovered_entity)

            arrow = Button(parent=self, model='cube', scale=.05, origin_z=-.5, color=color.green, collider='box')
            arrow.scale_z = distance(self, mouse.hovered_entity)
            arrow.look_at(mouse.hovered_entity)
            arrow.target = mouse.hovered_entity
            self.arrows.append(arrow)

            def break_connection(arrow=arrow):
                arrow.parent.connections.remove(arrow.target)
                arrow.parent.arrows.remove(arrow)
                destroy(arrow)

            arrow.on_click = break_connection

        # if held_keys['alt'] and key == 'left key down' and mouse.hovered_entity in self.arrows:
        #     self.connections.remove(mouse.hovered_entity.target)
        #     self.arrows.remove(mouse.hovered_entity)
        #     destroy(mouse.hovered_entity)


    def update(self):
        self.temp_arrow.enabled = self.connecting
        if self.connecting:
            self.color = color.green
            self.temp_arrow.look_at(mouse.world_point)
            self.temp_arrow.scale_z = distance(self.world_position, mouse.world_point)
        else:
            self.color = color.orange

def input(key):
    if held_keys['control'] and key == 'left mouse down':
        Node(world_position=mouse.world_point)


    if key == 'f':
        for e in scene.entities:
            if isinstance(e, Node):
                print(e.connections)
# def update():
#     print(mouse.world_point)



EditorCamera()





app.run()



from ursina import application  # noqa: I001
from ursina.entity import Entity
from ursina import curve
from ursina.ursinastuff import invoke
from ursina.destroy import destroy as _destroy
from ursina.string_utilities import print_warning
from ursina.scripts.property_generator import generate_properties_for_class
from pathlib import Path

from panda3d.core import Filename

from ursina.ursinastuff import DotDict
audio_groups = DotDict(
    music =     DotDict(volume_multiplier=1),
    ambient =   DotDict(volume_multiplier=1),
    sfx =       DotDict(volume_multiplier=1),
    dialogue =  DotDict(volume_multiplier=1),
)

audio_clip_cache = dict()

@generate_properties_for_class()
class Audio(Entity):

    volume_multiplier = .5  # master volume

    def __init__(self, sound_file_name='', volume=1, pitch=1, balance=0, loop=False, loops=1, autoplay=True, auto_destroy=False, group='sfx', **kwargs):
        super().__init__(**kwargs)
        # printvar(sound_file_name)
        self.clip = sound_file_name
        self.group = group
        if not self.clip:
            print_warning('missing audio clip:', sound_file_name)
            return

        self.volume = volume
        self.pitch = pitch
        self.balance = balance
        self.loop = loop
        if not loop:
            self.loops = loops
        self.autoplay = autoplay
        self.auto_destroy = auto_destroy


        if self.autoplay:
            self.play()

        if self.auto_destroy:
            invoke(self.stop, destroy=True, delay=self.length)


    def volume_setter(self, value):
        if not self._clip:  return
        self._volume = value
        group_volume_multiplier = audio_groups[self.group].volume_multiplier if self.group else 1
        self._clip.setVolume(value * Audio.volume_multiplier * group_volume_multiplier)

    def pitch_setter(self, value):
        if not self._clip:  return
        self._pitch = value
        self._clip.setPlayRate(value)

    def loop_setter(self, value):
        if not self._clip:  return
        self._loop = value
        self._clip.setLoop(value)

    def loops_setter(self, value):
        if not self._clip:  return
        self._loops = value
        self._clip.setLoopCount(value)


    def clip_setter(self, value):
        # if value in audio_clip_cache:
        #     self._clip = copy(audio_clip_cache[value])
        #     return

        if isinstance(value, Path):
            self._clip = loader.loadSfx(Filename.fromOsSpecific(str(value.resolve())))  # type: ignore
            return

        if isinstance(value, str):
            self.name = value

            if '.' in value:
                file_types = ('',)
            else:
                file_types = ('.ogg', '.wav', '.mp3')

            for folder in (application.asset_folder, application.internal_audio_folder):
                for suffix in file_types:
                    for f in folder.glob(f'**/{value}{suffix}'):
                        self.path = str(f.resolve())
                        self._clip = loader.loadSfx(Filename.fromOsSpecific(self.path))  # type: ignore
                        audio_clip_cache[value] = self._clip
                        # print('...loaded audio clip:', p, self._clip)
                        return

            self._clip = None
            print('no audio found with name:', value, 'supported formats: .ogg, .wav')
            return

        self._clip = value


    def length_getter(self):       # get the duration of the audio clip.
        return self.clip.length() if self.clip else 0

    def status_getter(self):
        if self.clip:
            return self.clip.status()

    def ready_getter(self):
        return 1 if self.clip and self.status > 0 else 0

    def playing_getter(self):
        return 1 if self.clip and self.status == 2 else 0

    def time_getter(self):
        return self.clip.get_time()
    def time_setter(self, value):
        self.clip.set_time(value)


    def balance_setter(self, value):    # pan the audio. should be a value between -.5 and .5. default: 0
        self._balance = value
        self.clip.setBalance(value*2)


    def play(self, start=0):
        if application.paused and not self.ignore_paused:
            return

        if self.clip:
            # print('play from:', start, self.clip)
            self.time = start
            self.clip.play()
        else:
            print(self, 'has no audio clip')

    def pause(self):
        if self.clip:
            self.paused_volume = self.volume
            self.paused_pitch = self.pitch
            self.paused_balance = self.balance
            self.temp_time = self.time
            self.pitch = 0
            self.stop(destroy=False)

    def resume(self):
        if self.clip and hasattr(self, 'temp_time'):
            self.clip = self.name
            self.volume = self.paused_volume
            self.pitch = self.paused_pitch
            self.balance = self.paused_balance
            self.play(self.temp_time)

    def stop(self, destroy=True):
        if self.clip:
            self.clip.stop()
        if destroy:
            _destroy(self)

    def fade(self, value, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt=True, ignore_paused=True):
        self.animate('volume', value=value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt, ignore_paused=ignore_paused)

    def fade_in(self, value=1, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt='finish', destroy_on_ended=False, ignore_paused=True):
        if duration+delay <= 0:
            self.volume = value
        else:
            self.animate('volume', value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt, ignore_paused=ignore_paused)
        if destroy_on_ended:
            _destroy(self, delay=delay+duration + .01)

    def fade_out(self, value=0, duration=.5, delay=0, curve=curve.in_expo, resolution=None, interrupt='finish', destroy_on_ended=True, ignore_paused=True):

        if duration+delay <= 0:
            self.volume = value
        else:
            self.animate('volume', value, duration=duration, delay=delay, curve=curve, resolution=resolution, interrupt=interrupt, ignore_paused=ignore_paused)
        if destroy_on_ended:
            _destroy(self, delay=delay+duration + .05)


if __name__ == '__main__':
    import random

    from ursina import Ursina

    app = Ursina()
    a = Audio('sine', loop=True, autoplay=True)

    a.volume = .5
    print('---', a.volume)
    # a = Audio('life_is_currency_wav', pitch=1)
    def input(key):
        if key == 'space':
            a = Audio('sine', pitch=random.uniform(.5,1), loop=True)
    # print(a.clip)
    # a.volume=0
    # b = Audio(a.clip)
    # a2 = Audio(clip=a.clip)
    # a2 = duplicate(a)
    # a2.clip = a.clip
    # a2.play()
    # print(a2.clip)
    # a.fade_out(delay=1)
    # DebugMenu(a)

    # def input(key):
    #     if key == 'f':
    #         a.fade_out(duration=4, curve=curve.linear)
    #
    # def update():
    #     print(a.time)



    app.run()



from ursina import *

app = Ursina()
random.seed(0)

ground = Entity(model='plane', origin=(-.5,0,-.5), scale=64, texture='quad_tree_terrain_heightmap')
ground.texture.filtering = None
EditorCamera()

import copy
# m = Prismatoid(base_shape=Quad(segments=0), path=((0,0,0),(0,1,0)), thicknesses=((1.3,1.3),(1,1)))
m = Prismatoid(base_shape=Circle(resolution=7, radius=1), path=((0,0,0),(0,1,0)), thicknesses=((1.3,1.3),(1,1)))
p=Entity()
def input(key):
    if key in ('space', 'space hold'):
        t = ground.texture
        exit = False

        for y in range(t.height):
            if exit:
                break
            for x in range(t.width):
                if exit:
                    break

                col = t.get_pixel(x,y)
                if col == color.black or col == color.red:
                    continue

                # find biggest square that fits
                for i in range(16):
                    if color.black in t.get_pixels((x,y), (x+i,y+i)):
                        break

                # printvar(i)
                i -= 1
                for _y in range(y, y+i):
                    for _x in range(x, x+i):
                        t.set_pixel(_x, _y, color.black)

                e = Entity(
                    parent=p,
                    model=copy.copy(m),
                    position=(x,0,y),
                    # origin=(-.5,0,-.5),
                    texture='white_cube',
                    scale=(i,4+(random.random()),i),
                    color = color.color(random.uniform(0,90), random.uniform(.2, .5), random.uniform(.9, 1.0)),
                    rotation_y = random.uniform(0,360),
                    )
                e.scale *= 1.1
                t.apply()
                exit = True

    if key == 'b':
        p.combine()
    #     p.model.colorize()

app.run()



import sys
sys.path.append("..")
from ursina import *

class ReplaceModelButton():

    def __init__(self):
        self.path = None

    def input(self, key):
        if key == 'left mouse down' and self.entity.hovered:
            if self.path:
                self.replace_models()

    @undoable
    def replace_models(self):
        original_models = list()
        target_entities = list()
        for s in scene.editor.selection:
            original_models.append(s.model)
            target_entities.append(s)

        for s in scene.editor.selection:
            try:
                s.model.removeAllChildren()
            except:
                pass
            if s.texture:
                self.temp_texture = os.path.basename(s.texture.getFullpath())
            s.model.removeNode()
            s.model = os.path.basename(self.path).split('.')[0]
            s.editor_collider = 'box'
            if self.temp_texture:
                s.texture = self.temp_texture

        scene.editor.inspector.update_inspector()

        yield 'replace model'
        for i in len(target_entities):
            # scene.editor.selections[i].model = original_models[i]
            # s.editor_collider = 'box'

            s = target_entities[i]

            try:
                s.model.removeAllChildren()
            except:
                pass
            if s.texture:
                self.temp_texture = os.path.basename(s.texture.getFullpath())
            s.model.removeNode()
            s.model = original_models[i]
            s.editor_collider = 'box'
            if self.temp_texture:
                s.texture = self.temp_texture
        scene.editor.inspector.update_inspector()



from ursina import *

window.vsync = False
app = Ursina()

application.asset_folder = application.asset_folder.parent
ground = Entity(model='plane', texture='heightmap_1', scale=512/20)

EditorCamera()

tex = ground.texture
verts = list()
vert_colors = list()
step = 10
t = time.time()
for z in range(tex.height // step):
    for x in range(tex.width // step):
        p = tex.get_pixel(x*step, z*step)
        h = int(color.brightness(p) * 10)
        # print(p)
        # z = y
        # verts += (
        #     (x,0,z), (x+1,0,z), (x+1,h,z), (x,0,z), (x+1,h,z), (x,h,z),
        #
        # )
        # verts.append(Vec3(x,h,z))
        # col = color.rgb(p[0], p[1], p[2])
        # col = color.tint(col, (z/255/2)-.2)
        # vert_colors += (lerp(p, color.random_color(), .1), )
        for y in range(0, h, 1):
            verts.append(Vec3(x, y, z))
            col = p.tint(random.uniform(-.01,.01))
            vert_colors.append(col)

print(time.time()-t)
scale = .2
e = Entity(model=Mesh(vertices=verts, colors=vert_colors, mode='point', render_points_in_3d=True, thickness=.75, static=True), scale=scale,
    texture='circle'
    )

for i in range(10):
    duplicate(e, x=i*20)
# e.model.set_render_mode_perspective(True)
# print(t.pixels)
# for row in t.pixels:
#     for c in row:
#         print((c[0] + c[1] + c[2]) / 3)
t = 0
# def update():
#     global t
#     t += time.dt
#     if t < .1:
#         return
#     t = 0
#     e.model.vertices = [v+Vec3(.05,0,0) for v in e.model.vertices]
#     e.model.generate()






app.run()



from ursina import Vec3, lerp_exponential_decay, time



class SmoothFollow:
    def __init__(self, target=None, offset=Vec3.zero, speed=8, rotation_speed=0, rotation_offset=Vec3.zero):
        self.target = target
        self.offset = Vec3(*offset)
        self.speed = speed
        self.rotation_speed = rotation_speed
        self.rotation_offset = rotation_offset


    def update(self):
        if not self.target:
            return

        self.entity.world_position = lerp_exponential_decay(self.entity.world_position, self.target.world_position+self.offset, time.dt*self.speed/4)

        if self.rotation_speed > 0:
            self.entity.world_rotation = lerp_exponential_decay(self.entity.world_rotation, self.target.world_rotation+self.rotation_offset, time.dt*self.rotation_speed/4)


if __name__ == '__main__':
    from ursina import Ursina, Entity, EditorCamera, color, held_keys
    app = Ursina()

    player = Entity(model='cube', color=color.orange)

    def update():
        player.x += held_keys['d'] * .1
        player.x -= held_keys['a'] * .1

    e = Entity(model='cube')
    sf = e.add_script(SmoothFollow(target=player, offset=(0,2,0)))

    def input(key):
        global sf
        if key == '1' and sf in e.scripts:
            e.scripts.remove(sf)

    EditorCamera()
    app.run()



import ast
import inspect
import textwrap
import typing

from ursina import *


class VertIn(typing.Generic[typing.TypeVar('T')]):
    pass
class FragIn(typing.Generic[typing.TypeVar('T')]):
    pass
class VertOut(typing.Generic[typing.TypeVar('T')]):
    pass
class FragOut(typing.Generic[typing.TypeVar('T')]):
    pass
class VertUni(typing.Generic[typing.TypeVar('T')]):
    pass
class FragUni(typing.Generic[typing.TypeVar('T')]):
    pass

py_to_glsl_precedence_dict = {
    ast.Add: 0,
    ast.Sub: 0,
    ast.Mult: 1,
    ast.Div: 1,
}

glsl_builtin_vars = {
    'gl_Position',
    'gl_FragCoord',
    'gl_FragColor',
}

def py_to_glsl_err(cls, node, error_msg, line_msg='here', indent=0):
    min_lineno = max(node.lineno - 1 - 3, 0)
    max_lineno = min(node.lineno - 1 + 3, len(cls.file_lines) - 1)
    msg = ''
    max_line_length = 50
    for lineno in range(min_lineno, max_lineno + 1):
        line = cls.file_lines[lineno]
        max_line_length = max(len(line), max_line_length)
        msg +=  line
        if lineno == node.lineno - 1:
            msg += '  <--------- ' + line_msg
        msg += '\n'
    sep = '-' * (max_line_length + 2)
    raise Exception(
        f'Shader error ({cls.__name__}:{node.lineno}): {error_msg}\n{sep}\n{msg}\n{sep}'
    )

def glsl_has_higher_precedence(a, b):
    return py_to_glsl_precedence_dict[type(a)] > py_to_glsl_precedence_dict[type(b)]

def py_constant_to_glsl(cls, node, indent=0):
    return f'{node.value}'

def py_name_to_glsl(cls, node, indent=0):
    return f'{node.id}'

def py_op_to_glsl(cls, node, indent=0):
    op_str = None

    if isinstance(node, ast.Add):
        op_str = '+'
    elif isinstance(node, ast.Sub):
        op_str = '-'
    elif isinstance(node, ast.Mult):
        op_str = '*'
    elif isinstance(node, ast.Div):
        op_str = '*'
    elif isinstance(node, ast.Eq):
        op_str = '=='
    elif isinstance(node, ast.Gt):
        op_str = '>'
    elif isinstance(node, ast.Lt):
        op_str = '<'

    return f'{op_str}'

def py_binop_to_glsl(cls, node, indent=0):
    op_str = py_op_to_glsl(cls, node.op)
    left_str = py_to_glsl(cls, node.left)
    right_str = py_to_glsl(cls, node.right)

    left_str_form = '{}'
    right_str_form = '{}'
    if isinstance(node.left, ast.BinOp) and glsl_has_higher_precedence(node.op, node.left.op):
        left_str_form = '({})'
    if isinstance(node.right, ast.BinOp) and glsl_has_higher_precedence(node.op, node.right.op):
        right_str_form = '({})'
    left_str = left_str_form.format(left_str)
    right_str = right_str_form.format(right_str)

    return f'{left_str} {op_str} {right_str}'


def py_ann_assign_to_glsl(cls, node, indent=0, is_global=False):
    var_name = py_to_glsl(cls, node.target)
    var_type = py_to_glsl(cls, node.annotation)
    var_value = None

    if hasattr(node, 'value') and node.value is not None:
        var_value = py_to_glsl(cls, node.value)

    prefix_str = ''
    if is_global:
        prefix_str = 'const'
        if isinstance(node.annotation, ast.Subscript):
            if node.annotation.value.id == 'VertIn' or node.annotation.value.id == 'FragIn':
                prefix_str = 'in'
            elif node.annotation.value.id == 'VertOut' or node.annotation.value.id == 'FragOut':
                prefix_str = 'out'
            elif node.annotation.value.id == 'VertUni' or node.annotation.value.id == 'FragUni':
                prefix_str = 'uniform'

    if var_value is None:
        if prefix_str != '':
            return '{}{} {} {};'.format(' ' * indent, prefix_str, var_type, var_name)
        else:
            return '{}{} {};'.format(' ' * indent, var_type, var_name)

    if prefix_str != '':
        return '{}{} {} {} = {};'.format(' ' * indent, prefix_str, var_type, var_name, var_value)

    return '{}{} {} = {};'.format(' ' * indent, var_type, var_name, var_value)

def py_subscript_to_glsl(cls, node, indent=0):
    value_str = py_to_glsl(cls, node.slice)

    return f'{value_str}'

def py_aug_assign_to_glsl(cls, node, indent=0):
    var_str = py_to_glsl(cls, node.target)
    op_str = py_op_to_glsl(cls, node.op)
    value_str = py_to_glsl(cls, node.value)

    return '{}{} {}= {};'.format(' ' * indent, var_str, op_str, value_str)

def py_assign_to_glsl(cls, node, indent=0):
    var_name = py_to_glsl(cls, node.targets[0])
    var_value = py_to_glsl(cls, node.value)

    return '{}{} = {};'.format(' ' * indent, var_name, var_value)

def py_compare_to_glsl(cls, node, indent=0):
    comp_str = py_op_to_glsl(cls, node.ops[0])

    left_str = py_to_glsl(cls, node.left)

    right_str = ''
    for c in node.comparators:
        right_str += py_to_glsl(cls, c)

    return f'{left_str} {comp_str} {right_str}'

def py_if_to_glsl(cls, node, indent=0):
    test_str = py_to_glsl(cls, node.test, indent=indent)

    body_str = ''
    for statement_node in node.body:
        body_str += py_to_glsl(cls, statement_node, indent=indent + 4) + '\n'

    return '{}if ({}) {{\n{}{}}}'.format(' ' * indent, test_str, body_str, ' ' * indent)

def py_return_to_glsl(cls, node, indent=0):
    return_str = py_to_glsl(cls, node.value)

    return '{}return {};'.format(' ' * indent, return_str)

def py_call_to_glsl(cls, node, indent=0):
    name_str = py_to_glsl(cls, node.func)

    args_str = ''
    for i, arg in enumerate(node.args):
        args_str += py_to_glsl(cls, arg)
        if i < len(node.args) - 1:
            args_str += ', '

    return '{}{}({})'.format(' ' * indent, name_str, args_str)

def py_atrribute_to_glsl(cls, node, indent=0):
    var_name_str = py_to_glsl(cls, node.value)

    return f'{" "*indent}{var_name_str}.{node.attr}'

def py_pass_to_glsl(cls, node, indent=0):
    return ''

def py_to_glsl(cls, node, indent=0):
    conversion_func = py_to_glsl_dict.get(type(node))
    if conversion_func is None:
        py_to_glsl_err(
            cls,
            node,
            f'Invalid Python AST for shader code in {cls.file_path} on line {node.lineno}.',
            line_msg=f'{type(node)}',
            indent=indent
        )
    return conversion_func(cls, node, indent=indent)

py_to_glsl_dict = {
    ast.Constant: py_constant_to_glsl,
    ast.Name: py_name_to_glsl,
    ast.BinOp: py_binop_to_glsl,
    ast.AnnAssign: py_ann_assign_to_glsl,
    ast.Subscript: py_subscript_to_glsl,
    ast.AugAssign: py_aug_assign_to_glsl,
    ast.Assign: py_assign_to_glsl,
    ast.Compare: py_compare_to_glsl,
    ast.If: py_if_to_glsl,
    ast.Return: py_return_to_glsl,
    ast.Call: py_call_to_glsl,
    ast.Attribute: py_atrribute_to_glsl,
    ast.Add: py_op_to_glsl,
    ast.Sub: py_op_to_glsl,
    ast.Mult: py_op_to_glsl,
    ast.Div: py_op_to_glsl,
    ast.Eq: py_op_to_glsl,
    ast.Gt: py_op_to_glsl,
    ast.Lt: py_op_to_glsl,
    ast.Pass: py_pass_to_glsl,
}

def pyshader(cls, glsl_version_number=150):
    class_file_string = ''
    class_file_path = inspect.getfile(cls)
    with open(class_file_path) as f:
        class_file_string = f.read()


    class_file_ast = ast.parse(class_file_string)

    glsl_vertex_source = f'#version {glsl_version_number} // Vertex Shader\n\n'
    glsl_fragment_source = f'#version {glsl_version_number} // Fragment Shader\n\n'

    has_vertex_shader = False
    has_fragment_shader = False

    cls.file_path = class_file_path
    cls.file_string = class_file_string
    cls.file_lines = cls.file_string.split('\n')
    for node in ast.walk(class_file_ast):
        if isinstance(node, ast.ClassDef) and node.name == cls.__name__:
            cls.lineno = node.lineno
            for class_child in ast.iter_child_nodes(node):
                if isinstance(class_child, ast.FunctionDef):
                    if class_child.name == 'vertex':
                        glsl_vertex_source += '\nvoid main() {\n'
                        has_vertex_shader = True
                        for statement in ast.iter_child_nodes(class_child):
                            if isinstance(statement, ast.arguments):
                                continue
                            glsl_string = py_to_glsl(cls, statement, indent=4)
                            glsl_vertex_source += glsl_string + '\n'
                        glsl_vertex_source += '}\n'
                    elif class_child.name == 'fragment':
                        glsl_fragment_source += '\nvoid main() {\n'
                        has_fragment_shader = True
                        for statement in ast.iter_child_nodes(class_child):
                            if isinstance(statement, ast.arguments):
                                continue
                            glsl_string = py_to_glsl(cls, statement, indent=4)
                            glsl_fragment_source += glsl_string + '\n'
                        glsl_fragment_source += '}\n'
                elif isinstance(class_child, ast.AnnAssign):
                    for_vert = False
                    for_frag = False
                    if isinstance(class_child.annotation, ast.Subscript):
                        if class_child.annotation.value.id.startswith('Vert'):
                            for_vert = True
                            for_frag = False
                        elif class_child.annotation.value.id.startswith('Frag'):
                            for_vert = False
                            for_frag = True
                    glsl_string = py_ann_assign_to_glsl(cls, class_child, is_global=True)
                    if for_vert:
                        glsl_vertex_source += glsl_string + '\n'
                    elif for_frag:
                        glsl_fragment_source += glsl_string + '\n'
                elif isinstance(class_child, ast.Assign):
                    py_assign_to_glsl(cls, class_child)
            del cls.lineno
            break
    del cls.file_lines
    del cls.file_string
    del cls.file_path

    if not has_vertex_shader:
        glsl_vertex_source = textwrap.dedent(
            f'''\
            //FALLBACK PROGRAM
            #version {glsl_version_number} // Vertex Shader\n
            uniform mat4 p3d_ModelViewProjectionMatrix;\n
            in vec4 p3d_Vertex;
            in vec2 p3d_MultiTexCoord0;\n
            out vec2 uv;\n
            void main() {{
                gl_Position = p3d_ModelViewProjectionMatrix * p3d_vertex;
                uv = p3d_MultiTexCoord0;
            }}
            '''
        )

    if not has_fragment_shader:
        glsl_fragment_source = textwrap.dedent(
            f'''\
            //FALLBACK PROGRAM
            #version {glsl_version_number} // Fragment Shader\n
            uniform sampler2D p3d_Texture0;
            uniform vec4 p3d_ColorScale;\n
            in vec2 uv;\n
            void main() {{
                gl_FragColor = texture(p3d_Texture0, uv) * p3d_ColorScale;
            }}
            '''
        )

    cls.glsl_vertex_source = glsl_vertex_source
    cls.glsl_fragment_source = glsl_fragment_source
    def shader_get_vertex_source():
        return cls.glsl_vertex_source
    def shader_get_fragment_source():
        return cls.glsl_fragment_source
    def shader_get_source():
        return cls.glsl_vertex_source + '\n//--------------------------\n\n' + cls.glsl_fragment_source
    cls.get_vertex_source = shader_get_vertex_source
    cls.get_fragment_source = shader_get_fragment_source
    cls.get_source = shader_get_source

    return cls


@pyshader
class MyShader:
    # class Vert:
    p3d_ModelViewProjectionMatrix: VertUni['mat4']
    p3d_Vertex: VertIn['vec4']
    p3d_MultiTexCoord0: VertIn['vec2']
    uv: VertOut['vec2']

    def vertex():
        gl_Position = p3d_ModelViewProjectionMatrix * p3d_Vertex
        uv = p3d_MultiTexCoord0;

    p3d_Texture0: FragUni['sampler2D']
    uv: FragIn['vec2']
    result: FragOut['vec4']

    def fragment():
        result = texture(p3d_Texture0, uv) * vec4(uv.x, uv.y, uv.x * uv.y, 1.0)




    # def fragment(p3d_Texture0:Texture, uv:Vec2) -> Vec4:
    #     result = texture(p3d_Texture0, uv) * vec4(uv.x, uv.y, uv.x * uv.y, 1.0)

# @pyshader
# class MyShaderExtended(MyShader):
#     def fragment():
#         return color.red



if __name__ == '__main__':
    app = Ursina()

    # #print(MyShader.get_source())
    print(MyShader.get_vertex_source())
    print(MyShader.get_fragment_source())

    my_shader = Shader(vertex=MyShader.get_vertex_source(), fragment=MyShader.get_fragment_source())

    box = Entity(model='cube', texture='brick', shader=my_shader)

    ed = EditorCamera()

    app.run()


from ursina import *


app = Ursina()


# e = Entity(model='terrain_1', scale=(20,5,40), texture='heightmap', texture_scale=(4,4))
e = Entity(model=Terrain('heightmap_1', skip=1), scale=(20,5,20), texture='heightmap', texture_scale=(2,2))
EditorCamera()
Sky()
app.run()



import builtins
from ursina.entity import Entity
from ursina.mesh import Mesh
from ursina.scene import instance as scene
from panda3d.core import CollisionTraverser, CollisionNode, CollisionHandlerQueue, CollisionRay
from ursina.vec3 import Vec3
from copy import copy
from ursina.hit_info import HitInfo
from ursina import ursinamath, color
from ursina.destroy import destroy



_line_model = Mesh(vertices=[Vec3(0,0,0), Vec3(0,0,1)], mode='line')

_raycaster = Entity(add_to_scene_entities=False)
_raycaster._picker = CollisionTraverser()  # Make a traverser
_raycaster._pq = CollisionHandlerQueue()  # Make a handler
_raycaster._pickerNode = CollisionNode('_raycaster')
_raycaster._pickerNode.set_into_collide_mask(0)
_raycaster._pickerNP = _raycaster.attach_new_node(_raycaster._pickerNode)
_raycaster._picker.addCollider(_raycaster._pickerNP, _raycaster._pq)
_ray = CollisionRay()
_ray.setOrigin(Vec3(0,0,0))
_ray.setDirection(Vec3(0,0,1))
_raycaster._pickerNode.addSolid(_ray)


def raycast(origin, direction:Vec3=(0,0,1), distance=9999, traverse_target:Entity=scene, ignore:list=None, debug=False, color=color.white):
    if not ignore:
        ignore = []

    _raycaster.position = origin
    _raycaster.look_at(_raycaster.position + direction)

    if debug:
        temp = Entity(position=origin, model=copy(_line_model), scale=Vec3(1,1,min(distance,9999)), color=color, add_to_scene_entities=False)
        temp.look_at(_raycaster.position + direction)
        destroy(temp, 1/30)

    _raycaster._picker.traverse(traverse_target)      #HALF!

    if _raycaster._pq.get_num_entries() == 0:
        _raycaster.hit = HitInfo(hit=False, distance=distance)
        return _raycaster.hit


    _raycaster._pq.sort_entries()
    entries = _raycaster._pq.getEntries()
    entities = [e.get_into_node_path().parent for e in entries]

    entries = [        # filter out ignored entities
        e for i, e in enumerate(entries)
        if entities[i] in scene.collidables
        and entities[i] not in ignore
        and ursinamath.distance(_raycaster.world_position, e.get_surface_point(builtins.render)) <= distance
        ]

    if len(entries) == 0:
        return HitInfo(hit=False)

    _raycaster.collision = entries[0]
    nP = _raycaster.collision.get_into_node_path().parent
    point = Vec3(*_raycaster.collision.get_surface_point(nP))
    world_point = Vec3(*_raycaster.collision.get_surface_point(builtins.render))

    hit_info = HitInfo(hit=True)
    hit_info.entities = [e.get_into_node_path().parent.getPythonTag('Entity') for e in entries]
    hit_info.entity = hit_info.entities[0]

    hit_info.point = point
    hit_info.world_point = world_point
    hit_info.distance = ursinamath.distance(_raycaster.world_position, hit_info.world_point)

    hit_info.normal = Vec3(*_raycaster.collision.get_surface_normal(_raycaster.collision.get_into_node_path().parent).normalized())
    hit_info.world_normal = Vec3(*_raycaster.collision.get_surface_normal(builtins.render).normalized())

    return hit_info


if __name__ == '__main__':
    from ursina import *
    from ursina import Ursina, Entity, held_keys, time, duplicate, camera, EditorCamera
    app = Ursina()

    '''
    Casts a ray from *origin*, in *direction*, with length *distance* and returns
    a HitInfo containing information about what it hit. This ray will only hit entities with a collider.

    Use optional *traverse_target* to only be able to hit a specific entity and its children/descendants.
    Use optional *ignore* list to ignore certain entities.
    Setting debug to True will draw the line on screen.

    Example where we only move if a wall is not hit:
    '''


    class Player(Entity):

        def update(self):
            self.direction = Vec3(
                self.forward * (held_keys['w'] - held_keys['s'])
                + self.right * (held_keys['d'] - held_keys['a'])
                ).normalized()  # get the direction we're trying to walk in.

            origin = self.world_position + (self.up*.5) # the ray should start slightly up from the ground so we can walk up slopes or walk over small objects.
            hit_info = raycast(origin , self.direction, ignore=(self,), distance=.5, debug=False)
            if not hit_info.hit:
                self.position += self.direction * 5 * time.dt
            else:
                print(hit_info.entity)

    Player(model='cube', origin_y=-.5, color=color.orange)
    wall_left = Entity(model='cube', collider='box', scale_y=3, origin_y=-.5, color=color.azure, x=-4)
    wall_right = duplicate(wall_left, x=4)
    camera.y = 2
    app.run()

    # # test
    # # breakpoint()
    # d = Entity(parent=scene, position=(0,0,2), model='cube', color=color.orange, collider='box', scale=2)
    # e = Entity(model='cube', color=color.lime)
    #
    # camera.position = (0, 15, -15)
    # camera.look_at(e)
    # # camera.reparent_to(e)
    # speed = .01
    # rotation_speed = 1
    # intersection_marker = Entity(model='cube', scale=.2, color=color.red)
    #
    # def update():
    #     e.position += e.forward * held_keys['w'] * speed
    #     e.position += e.left * held_keys['a'] * speed
    #     e.position += e.back * held_keys['s'] * speed
    #     e.position += e.right * held_keys['d'] * speed
    #
    #     e.rotation_y -= held_keys['q'] * rotation_speed
    #     e.rotation_y += held_keys['e'] * rotation_speed
    #
    #     ray = raycast(e.world_position, e.forward, 3, debug=True)
    #     # ray = raycast(e.world_position, e.forward, 3, debug=True)
    #     # ray = boxcast(e.world_position, e.right, 3, debug=True)
    #     # print(ray.distance, ray2.distance)
    #     intersection_marker.world_position = ray.world_point
    #     intersection_marker.visible = ray.hit
    #     if ray.hit:
    #         d.color = color.azure
    #     else:
    #         d.color = color.orange

    # ray = raycast(e.world_position, e.forward, 3, debug=True)

    EditorCamera()
    app.run()



from direct.showbase.ShowBase import ShowBase
# from ShaderCombiner import ShaderCombiner
from direct.gui.DirectSlider import DirectSlider
from panda3d.core import GeomPatches, GeomVertexFormat, GeomVertexWriter, GeomVertexData, Geom, GeomNode


from os.path import join
from panda3d.core import Shader

from os import makedirs
from os.path import isdir


class ShaderCombiner:

    __tmp_path = "shader_dump"

    @classmethod
    def setTempPath(self, p):
        self.__tmp_path = p

    @classmethod
    def ensureTempPath(self):
        if isdir(self.__tmp_path):
            return
        try:
            makedirs(self.__tmp_path)
        except Exception as msg:
            print ("Could not create shader dir at", self.__tmp_path, ":", msg)

    @classmethod
    def getShaderFileContents(self, filename):
        with open(filename) as f:
            return f.read()

    @classmethod
    def writeShaderFileContents(self, filename, content, append=False):
        with open(filename, 'a' if append else 'w') as f:
            f.write(content)

    @classmethod
    def parseShaderContent(self, content):
        content = content.replace("\r", "")
        return content

    @classmethod
    def createGlslShader(self, files, has_geometry, name, glslversion=130, externals=None, defines=None, has_tesselation=False):

        dbg = "Creating shader '" + str(name) + "' with glslversion " + str(glslversion)
        if has_geometry:
            dbg += " (geometry)"
        if has_tesselation:
            dbg += " (tesselated)"
        print(dbg)


        self.ensureTempPath()

        if type(files) == str:
            files = [files]

        if type(files) != list:
            files = list(files)

        files = [] + files

        header = "//GLSL\n"
        header += "#version " + str(glslversion) + "\n"


        if defines is not None:
            for dname, dvalue in defines.items():
                header += "#define " + dname + " " + str(dvalue) + "\n"

        content_vshader = header
        content_fshader = header
        content_gshader = header
        content_tessControl = header
        content_tessEval    = header


        content_vshader += "#define _VERTEX_\n\n#define VARIABLE out\n\n"

        if has_geometry:
            content_fshader += "#extension GL_ARB_fragment_layer_viewport : enable\n"

        content_fshader += "#define _FRAGMENT_\n\n#define VARIABLE in\n\n"

        if has_geometry:
            content_gshader += "#extension GL_EXT_geometry_shader4 : enable\n"
            content_gshader += "#extension GL_ARB_fragment_layer_viewport : enable\n"
            content_gshader += "#define _GEOMETRY_\n\n"

        if has_tesselation:
            content_tessControl += "#define _TESSCONTROL_\n"
            content_tessEval    += "#define _TESSEVAL_\n"




        combined_content = ""

        for f in files:
            c = self.getShaderFileContents(f)
            combined_content += self.parseShaderContent(c) + "\n\n"

        if externals is not None:
            for eid, source in externals.items():
                repl = '#EXTERNAL("' + eid + '")'
                content = self.getShaderFileContents(source)
                combined_content = combined_content.replace(repl, content)

        content_vshader += combined_content
        content_fshader += combined_content
        content_gshader += combined_content
        content_tessControl += combined_content
        content_tessEval    += combined_content

        f_vertex = join(self.__tmp_path, name + "_vertex.glsl")
        f_fragment = join(self.__tmp_path, name + "_fragment.glsl")
        f_geometry = join(self.__tmp_path, name + "_geometry.glsl")

        f_tessControl = join(self.__tmp_path, name + "_tesselationControl.glsl")
        f_tessEval    = join(self.__tmp_path, name + "_tesselationEvaluation.glsl")


        self.writeShaderFileContents(f_vertex, content_vshader)
        self.writeShaderFileContents(f_fragment, content_fshader)

        if has_geometry:
            self.writeShaderFileContents(f_geometry, content_gshader)

        if has_tesselation:
            self.writeShaderFileContents(f_tessControl, content_tessControl)
            self.writeShaderFileContents(f_tessEval, content_tessEval)

        args = [f_vertex, f_fragment]

        if has_geometry:
            args.append(f_geometry)

        if has_tesselation:
            if not has_geometry:
                args.append("")
            args.append(f_tessControl)
            args.append(f_tessEval)


        shader = Shader.load(Shader.SLGLSL, *args)
        return shader

class MyApp(ShowBase):

    def __init__(self):
        ShowBase.__init__(self)

        # Camera stuff
        base.disableMouse()
        base.camera.setPos(0, -3, 0)
        base.camera.lookAt(0, 0, 0)

        # Toggle wireframe so we can see the vertices
        base.toggleWireframe()

        # For reloading the shaders
        self.accept("r", self.loadShader)

        # Create the patch
        form = GeomVertexFormat.getV3()
        vdata = GeomVertexData("vertices", form, Geom.UHStatic)

        vertexWriter = GeomVertexWriter(vdata, "vertex")
        vertexWriter.addData3f(0, 0, 0)
        vertexWriter.addData3f(1, 0, 0)
        vertexWriter.addData3f(1, 0, 1)
        vertexWriter.addData3f(0, 0, 1)

        patches = GeomPatches(3, Geom.UHStatic)
        patches.addVertex(0)
        patches.addVertex(1)
        patches.addVertex(3)
        patches.closePrimitive()

        patches.addConsecutiveVertices(1, 3)
        patches.closePrimitive()

        gm = Geom(vdata)
        gm.addPrimitive(patches)

        gn = GeomNode("model")
        gn.addGeom(gm)
        self.tesselatedQuad = render.attachNewNode(gn)
        self.tesselatedQuad.setPos(-0.5, 0, -0.5)

        # Load tesselation shader
        self.loadShader()

        # Set initial tesselation factors
        self.tesselatedQuad.setShaderInput("TessLevelInner", 5.0)
        self.tesselatedQuad.setShaderInput("TessLevelOuter", 5.0)

        # Sliders to modify the tesselation factors
        self.tessInnerSlider = DirectSlider(range=(1, 20), value=8.0, pageSize=1.0, pos=(
            0, 0, 0.8), scale=(1.1, 1, 0.4), command=self.setInnerTess)
        self.tessOuterSlider = DirectSlider(range=(1, 20), value=8.0, pageSize=1.0, pos=(
            0, 0, 0.7), scale=(1.1, 1, 0.4), command=self.setOuterTess)

    def setInnerTess(self):
        self.tesselatedQuad.setShaderInput(
            "TessLevelInner", float(self.tessInnerSlider['value']))

    def setOuterTess(self):
        self.tesselatedQuad.setShaderInput(
            "TessLevelOuter", float(self.tessOuterSlider['value']))

    def loadShader(self):
    # #     # This simply loads the tesselation shaders
    # #     sha = ShaderCombiner.createGlslShader(
    # #         ["shader.glsl"], False, "shader", glslversion=400, externals=None, defines=None, has_tesselation=True)
        from tesselation_test import tesselation_shader
    #     self.tesselatedQuad.setShader(tesselation_shader._shader)


app = MyApp()
app.run()



from ursina import *

'''
nineslice mesh:

11-10--15--14
|   |   |   |
8---9--12--13
|   |   |   |
3---2---7---6
|   |   |   |
0---1---4---5

'''


class NineSlice(Mesh):
    def __init__(self, entity_scale=Vec2.one, radius=.5, scale_multiplier=1):
        verts = [Vec2(*e)*scale_multiplier for e in (
            (-.5,-.5), (-.5,-.5), (-.5,-.5), (-.5,-.5),
            (.5,-.5), (.5,-.5), (.5,-.5), (.5,-.5),
            (-.5,.5), (-.5,.5), (-.5,.5), (-.5,.5),
            (.5,.5), (.5,.5), (.5,.5), (.5,.5),
            )]
        if isinstance(entity_scale, Vec3):
            entity_scale = entity_scale.xy
        elif isinstance(entity_scale, float | int):
            entity_scale = Vec2(entity_scale)
        elif isinstance(entity_scale, tuple | list):
            entity_scale = Vec2(*entity_scale)

        if entity_scale.x > entity_scale.y:
            aspect_ratio = (entity_scale.x / entity_scale.y)
            print('aspect_ratio', aspect_ratio, entity_scale)
            for idx in (1,2,9,10):
                verts[idx].x += radius / aspect_ratio
            for idx in (4,7,12,15):
                verts[idx].x -= radius / aspect_ratio
            for idx in (3,2,7,6):
                verts[idx].y += radius
            for idx in (8,9,12,13):
                verts[idx].y -= radius

        else:
            aspect_ratio = (entity_scale.y / entity_scale.x)
            for idx in (1,2,9,10):
                verts[idx].x += radius
            for idx in (4,7,12,15):
                verts[idx].x -= radius
            for idx in (3,2,7,6):
                verts[idx].y += radius / aspect_ratio
            for idx in (8,9,12,13):
                verts[idx].y -= radius / aspect_ratio


        super().__init__(
            vertices=[Vec3(*p,0) for p in verts],
            triangles=(
                (0,1,2,3),
                (1,4,7,2),
                (4,5,6,7),
                (3,2,9,8,),
                (2,7,12,9),
                (7,6,13,12),
                (8,9,10,11),
                (9,12,15,10),
                (12,13,14,15)
                ),
            uvs=(
                (0,0), (.5,0), (.5,.5), (0,.5),
                (.5,0), (1,0), (1,.5), (.5,.5),
                (0,.5), (.5,.5), (.5,1), (0,1),
                (.5,.5), (1,.5), (1,1), (.5,1)
                ),
            )



if __name__ == '__main__':
    app = Ursina()
    m = NineSlice(entity_scale=Vec3(2,1,1),
        # scale_multiplier=1.4,   # account for whitespace and drop shadow in the texture
        )

    window.color = color._0

    Entity(parent=camera.ui, model='wireframe_quad', scale=Vec2(2,1)*.1, color=color.red)
    # for i, (nineslice_texture, scale_multiplier) in enumerate(
    #     zip(
    #         ('nineslice_rainbow', 'nineslice_double', 'circle'),
    #         (1.4, 1.4, 1), strict=True
    #     )):
    #     b = Draggable(
    #         model=NineSlice(entity_scale=Vec3(2,1,1), scale_multiplier=scale_multiplier),
    #         scale=Vec2(2,1)*.1,
    #         texture=nineslice_texture,
    #         text='Next',
    #         color=color.white,
    #         text_color=color.dark_gray,
    #         text_origin=(0,0),
    #         highlight_color=hsv(210,.1,1),
    #         pressed_scale=.9,
    #         pressed_color=hsv(180,.2,.7),
    #         # wireframe=True,
    #         y=-i*.2
    #         )

    # button1 = Entity(text="Hi !!", model=NineSlice(2, 3), scale=(.2,.3), position=(-0.2, -0.3), texture="nineslice_rainbow", color=color.white)
    button2 = Draggable(parent=scene, model=NineSlice((1,1)), scale=1, position=(0,0), origin=(-.5,-.5), texture="nineslice_rainbow", color=color.white)
    button2 = Draggable(parent=scene, model=NineSlice((2,1)), scale=(2,1), position=(0,1), origin=(-.5,-.5), texture="nineslice_rainbow", color=color.white, wireframe=False)
    button2 = Draggable(parent=scene, model=NineSlice((1,3)), scale=(1,3), position=(1,-2), origin=(-.5,-.5), texture="nineslice_rainbow", color=color.white, wireframe=False)
    button2 = Draggable(parent=scene, model=NineSlice((3,1)), scale=(3,1), position=(-2,-1), origin=(-.5,-.5), texture="nineslice_rainbow", color=color.white, wireframe=False)
    # button3 = Entity(text="As you can see, the texture goes \nwell for every button size.", model=NineSlice(0.4, 0.4), position=(0, 0.1), texture="nine_sliced", color=color.white)
    EditorCamera()
    Entity(model=Grid(8,8), scale=8)


    app.run()


from ursina import *


app = Ursina()

e = Entity(model=Terrain('loddefjord_height_map', skip=8), texture='loddefjord_color', scale=100, scale_y=50)
Sky(rotation_y=125)
e.model.save('loddefjord_terrain')
scene.fog_color = color.gray
scene.fog_density = .01

from ursina.prefabs.first_person_controller import FirstPersonController
# fpc = FirstPersonController(speed=10)
EditorCamera()
window.exit_button.visible = False
window.fps_counter.enabled = False
mouse.visible = False



# from ursina .shaders import camera_vertical_blur_shader
# camera.shader = camera_vertical_blur_shader
# camera.set_shader_input('blur_size', .0)
#
# camera.blur_amount = 0
# def update():
#     camera.set_shader_input("blur_size", camera.blur_amount)
#
# t = Text('It was like a dream', enabled=False, scale=3, origin=(0,0), color=color.dark_text)
#
# def input(key):
#     if key == 'space':
#         camera.animate('blur_amount', .6, duration=5)
#         invoke(t.appear, speed=.1, delay=5.2)

# window.size /= 2


app.run()



from ursina import *


class ProjectBrowser(Entity):

    def __init__(self):
        super().__init__()
        self.name = 'project_browser'

        # window.size = (1920 * .6, 1080 * .6)
        # window.name = self.name
        # window.color = color.color(0, 0, .9)
        # color.text_color = color.color(0, 0, .1)
        # camera.ui.scale *= .95

        self.bg = Entity()
        self.bg.parent = application.ui
        self.bg.model = 'quad'
        self.bg.scale_x *= camera.aspect_ratio
        self.bg.texture = 'project_browser_bg'
        self.bg.color = color.gray

        # self.title = Text()
        self.title = Text()
        self.title.parent = application.ui
        self.title.text = 'Choose Project'
        # self.title.align = 'center'
        self.title.scale /= 4
        self.title.position = (-.5 * camera.aspect_ratio + .05, .425)

        self.list = Entity()
        self.list.parent = application.ui
        self.list.x = -.5 * camera.aspect_ratio + .05
        # self.list.scale *= .9

        # from project_paths import paths
        self.paths = ('new', 'open', 'test0', 'test1', 'test2', 'test3')
        self.index = 0
        if len(self.paths) > 0:
            self.index = 1



        for i in range(len(self.paths)):
            self.b = Button()
            s = self.b.add_script(ProjectBrowserButton())
            s.index = i
            self.b.parent = self.list
            self.b.collider = 'box'
            self.b.origin = (-.5, .5)
            self.b.x = 0 + (i * .34)
            self.b.scale = (.33 * 1, .33)
            self.b.color = color.color(i * 20, .2, 1, .5)
            if i == 0:
                self.b.color = color.color(0, 0, .75, .5)
            # self.b.texture = 'white_cube'

            t = Text()
            t.text = 'Name' + str(i)
            t.parent = self.b.model
            t.scale *= .5
            t.position = (-.5, .57)
            if i == 0:
                t.text = '+'
                t.align = 'center'
                t.scale *= 3
                t.position = (0, 0)

            if i > 0:
                t = Text()
                t.text = self.paths[i]
                t.parent = self.b.model
                t.scale *= .25
                t.position = (-.5, -.55)


    def input(self, key):
        if len(self.paths) > 5:
            if key == 'scroll up':
                self.list.x += .1
            if key == 'scroll down':
                self.list.x -= .1

            self.list.x = clamp(self.list.x, (-len(self.paths) * .34) + .84 , -.84)


class ProjectBrowserButton():

    def input(self, key):
        if self.entity.hovered:
            if (key == 'left mouse down'
            or key == 'space'
            or key == 'enter'
            and self.index > 0):
                self.load_project(application.paths[self.index])

    def on_mouse_enter(self):
        application.index = self.index

    def load_project(self, path):
        print('loading project:', path)
        if os.path.isdir(path):
            print('yay')
            subprocess.call('python.exe ' + path + 'main.py')
            # run process python.exe, path + main.py, editor=True
            os._exit(0)     # force exit
        else:
            print('project not found at:', path)

if __name__ == '__main__':
    app = main.Ursina()
    load_scene('project_browser')
    # t = ProjectBrowser()
    app.run()



from ursina import *

app = Ursina()

title = Text('''position, rotation, scale and parenting demo''', position=(-.85, .475), scale=1.5)

entity = Entity(model='cube', texture='white_cube')
child = Entity(parent=entity, x=1, model='cube', scale=.5, texture='white_cube', color=color.azure)
spacing = .04
ui_parent = Entity(parent=camera.ui, scale=.8, x=-.85)

for y, e in enumerate((entity, child)):
    for i, name in enumerate(('x', 'y', 'z')):
        slider = Slider(-1, 1, default=getattr(e, name), text=name, dynamic=True, position=(.1, .4-(i*spacing)-y*.5), parent=ui_parent)
        def on_slider_changed(e=e, slider=slider, attr_name=name):
            setattr(e, attr_name, slider.value)
        slider.on_value_changed = on_slider_changed

        name = 'rotation_' + name
        slider = Slider(0, 360, default=getattr(e, name), text=name, dynamic=True, position=(.1, .4-(i*spacing)-(spacing*3)-y*.5), parent=ui_parent)
        def on_slider_changed(e=e, slider=slider, attr_name=name):
            setattr(e, attr_name, slider.value)
        slider.on_value_changed = on_slider_changed

        name = 'scale_' + name.split('_')[1]
        slider = Slider(.1, 3, default=getattr(e, name), text=name, dynamic=True, position=(.1, .4-(i*spacing)-(spacing*6)-y*.5), parent=ui_parent)
        def on_slider_changed(e=e, slider=slider, attr_name=name):
            setattr(e, attr_name, slider.value)
        slider.on_value_changed = on_slider_changed

t = Text('''white cube:''', position=(0, .4+spacing*1.5), parent=ui_parent)
t.create_background(.025, 0.0125)
t = Text('''blue cube (parented to the white cube):''', position=(0, -spacing), parent=ui_parent)
t.create_background(.025, 0.0125)

EditorCamera()

app.run()



from ursina import *


class Inventory(Entity):
    def __init__(self, width=5, height=8, **kwargs):
        super().__init__(
            parent = camera.ui,
            model = Quad(radius=.015),
            texture = 'white_cube',
            texture_scale = (width, height),
            scale = (width*.1, height*.1),
            origin = (-.5,.5),
            position = (-.3,.4),
            color = color.hsv(0, 0, .1, .9),
            )

        self.width = width
        self.height = height

        for key, value in kwargs.items():
            setattr(self, key, value)


    def find_free_spot(self):
        for y in range(self.height):
            for x in range(self.width):
                grid_positions = [(int(e.x*self.texture_scale[0]), int(e.y*self.texture_scale[1])) for e in self.children]
                print(grid_positions)

                if not (x, -y) in grid_positions:
                    print('found free spot:', x, y)
                    return x, y


    def append(self, item, x=0, y=0):
        print('add item:', item)

        if len(self.children) >= self.width*self.height:
            print('inventory full')
            error_message = Text('<red>Inventory is full!', origin=(0,-1.5), x=-.5, scale=2)
            destroy(error_message, delay=1)
            return

        x, y = self.find_free_spot()

        icon = Draggable(
            parent = self,
            model = 'quad',
            texture = item,
            color = color.white,
            scale_x = 1/self.texture_scale[0],
            scale_y = 1/self.texture_scale[1],
            origin = (-.5,.5),
            x = x * 1/self.texture_scale[0],
            y = -y * 1/self.texture_scale[1],
            z = -1,
            )
        name = item.replace('_', ' ').title()

        if random.random() < .25:
            icon.color = color.gold
            name = '<orange>Rare ' + name

        icon.tooltip = Tooltip(name)
        icon.tooltip.background.color = color.hsv(0,0,0,.8)


        def drag():
            icon.org_pos = (icon.x, icon.y)
            icon.z = -2   # ensure the dragged item overlaps the rest

        def drop():
            icon.x = int((icon.x + (icon.scale_x/2)) * self.width) / self.width
            icon.y = int((icon.y - (icon.scale_y/2)) * self.height) / self.height
            icon.z = -1

            # if outside, return to original position
            if icon.x < 0 or icon.x >= 1 or icon.y > 0 or icon.y <= -1:
                icon.position = (icon.org_pos)
                return

            # if the spot is taken, swap positions
            for c in self.children:
                if c == icon:
                    continue

                if c.x == icon.x and c.y == icon.y:
                    print('swap positions')
                    c.position = icon.org_pos

        icon.drag = drag
        icon.drop = drop



if __name__ == '__main__':
    app = Ursina()

    inventory = Inventory()

    def add_item():
        inventory.append(random.choice(('bag', 'bow_arrow', 'gem', 'orb', 'sword')))

    add_item()
    add_item()
    add_item_button = Button(
        scale = (.1,.1),
        x = -.5,
        color = color.lime.tint(-.25),
        text = '+',
        tooltip = Tooltip('Add random item'),
        on_click = add_item
        )
    background = Entity(parent=camera.ui, model='quad', texture='shore', scale_x=camera.aspect_ratio, z=1)

    Cursor(texture='cursor', scale=.1)
    mouse.visible = False

    app.run()



from ursina import *
from copy import copy


app = Ursina()

cube = load_model('cube', use_deepcopy=True)
cube.colorize(color.smoke, color.light_gray, color.dark_gray, color.white, color.smoke, color.light_gray)

# ground = Entity(model='plane', texture='white_cube', origin=(-.5,0,-.5), scale=20, texture_scale=(20,20), rotation_y=45)
ground = Entity(model=Plane((20,20)), scale=400, texture='grass', texture_scale=(10,10))
Sky()


# e = Entity(model=copy(cube), origin_y=-.5, scale=172, scale_y=10)
# e = Entity(model=copy(cube), origin_y=-.5, scale=126, scale_y=20)
#
# def input(key):
#     if key == 'space':
#         e.scale_x -= 2
#         print(e.scale_x)

t = time.time()
def make_maze():
    maze_parent = Entity(scale=10, position=(-100, 0, -100))
    goal_post = Entity(
        model=Cylinder(resolution=24),
        origin=(.25,0,.25),
        scale=16,
        scale_y=32,

        )
    # return
    i = 0
    for z in range(20):
        for x in range(20):
            if x > 6 and x < 13 and z > 6 and z < 13:
                continue
            #
            # if x < 8:
            #     x_lvl = x//2
            # elif x > 13:
            #     x_lvl = 20-x
            #
            # z_lvl = 10
            #
            # level = min(x_lvl, z_lvl)

            invoke(Entity,
                parent=maze_parent,
                model=copy(cube),
                # texture='white_cube',
                # collider='box',
                # color=color.light_gray,
                scale_x=.25,
                scale_z=1.65,
                origin_y=-.5,
                position=(x+random.uniform(-.01,.01), 0+random.uniform(-.01,.01), z+random.uniform(-.01,.01)),
                rotation_y=random.choice([45,-45]),
                delay=i
                )
            # invoke(eval, code, delay=i)
            i += .005

def input(key):
    if key == 'space':
        make_maze()

print(time.time() - t)

Entity(model='cube', color=color.red)
# camera.position = (0,400,-400)
# camera.look_at((0,0,0))
camera.z = -800
EditorCamera(rotation_x=45)

# from ursina.prefabs.first_person_controller import FirstPersonController
# player = FirstPersonController(speed=10)
#

# from ursina .shaders import camera_vertical_blur_shader
# camera.shader = camera_vertical_blur_shader
# camera.set_shader_input('blur_size', .05)
# slider = ThinSlider(max=.1, dynamic=True, position=(-.25, -.45))
#
# def set_blur():
#     camera.set_shader_input("blur_size", slider.value)
#
# slider.on_value_changed = set_blur

# player.enabled = False
# camera.position = (0,100,-200)
# camera.look_at((0,0,0))

# camera.animate_position(player.position + player.up, duration=4, curve=curve.linear)
# camera.animate_rotation(player.rotation, duration=10, curve=curve.out_expo)
# invoke(setattr, player, 'enabled', True, delay=4)

app.run()




from math import ceil, floor

from ursina.array_tools import Array2D, chunk_list, enumerate_2d
from ursina.ursinamath import clamp


def chunk_mesh(mesh, num_chunks, chunk_size=1/8):
    if not hasattr(mesh, 'vertices'):
        raise Exception(f'{mesh} has no vertices')

    chunks = Array2D(width=num_chunks[0], height=num_chunks[1])
    for (x,z), _value in enumerate_2d(chunks):
        chunks[x][z] = []

    polys = list(chunk_list(mesh.generated_vertices, 3))

    for i, tri in enumerate(polys):
        if chunk_size >= 1:
            min_x = floor(min(v.x for v in tri) / chunk_size)
            min_y = floor(min(v.z for v in tri) / chunk_size)
            max_x = ceil(max(v.x for v in tri) / chunk_size)
            max_y = ceil(max(v.z for v in tri) / chunk_size)
        else:
            min_x = floor(min(v.x for v in tri) / chunk_size)
            min_y = floor(min(v.z for v in tri) / chunk_size)
            max_x = ceil(max(v.x for v in tri) / chunk_size)
            max_y = ceil(max(v.z for v in tri) / chunk_size)

        min_x = clamp(min_x, 0, chunks.width-1)
        max_x = clamp(max_x, 0, chunks.width-1)
        min_y = clamp(min_y, 0, chunks.height-1)
        max_y = clamp(max_y, 0, chunks.height-1)

        # print(min_x, max_x, min_y, max_y)
        for x in range(min_x, max_x+1):
            for y in range(min_y, max_y+1):
                # print('append tri', i, 'to chunk', x, y)
                chunks[x][y].append(tri)

    return chunks

if __name__ == '__main__':
    from ursina import EditorCamera, Entity, Grid, Mesh, Terrain, Ursina, Vec2, Vec3, color, flatten_list
    app = Ursina()

    reference = Entity(x=-1, model=Terrain('heightmap_1', skip=4), texture='grass', texture_scale=(3,3), origin=(-.5,0,-.5), wireframe=True, color=color.green)
    target_mesh = reference.model
    target_mesh.vertices = [v+Vec3(.5,0,.5) for v in target_mesh.vertices]
    chunk_size = 1/8
    num_chunks = Vec2(8,8)
    chunks = chunk_mesh(target_mesh, num_chunks, chunk_size)
    grid = Entity(x=-1, model=Grid(*num_chunks), rotation_x=90, color=color.red, scale=num_chunks*chunk_size, origin=(-.5,-.5))

    for (x,z), value in enumerate_2d(chunks):
        if not value:
            chunks[x][z] = Entity(model='cube', color=color.random_color(), position=Vec3(x-1,0,z)*chunk_size, origin=(-.5,-.5,-.5), scale=chunk_size)
            continue
        chunks[x][z] = Entity(x=-1, model=Mesh(vertices=flatten_list(value)), color=color.random_color())


    # target_mesh = load_model('tower_collider.ursinamesh')
    # reference = Entity(model=target_mesh, wireframe=True)
    # chunk_size = 8
    # num_chunks = Vec2(5,7)
    # chunks = chunk_mesh(target_mesh, num_chunks, chunk_size)
    # grid = Entity(model=Grid(*num_chunks), rotation_x=90, color=color.red, scale=num_chunks*chunk_size, origin=(-.5,-.5))

    # for (x,z), value in enumerate_2d(chunks):
    #     if not value:
    #         chunks[x][z] = Entity(model='wireframe_cube', color=color.random_color(), position=Vec3(x,0,z)*chunk_size, origin=(-.5,-.5,-.5), scale=chunk_size)
    #         continue
    #     chunks[x][z] = Entity(model=Mesh(vertices=flatten_list(value)), color=color.random_color())

    EditorCamera()

    app.run()